"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import click
import sys
import json
import questionary
from hcs_core.ctxp import util, panic, profile, cli_processor


@click.group(name="profile", cls=cli_processor.LazyGroup)
def profile_cmd_group():
    """Commands for profile."""


@profile_cmd_group.command()
def list():
    """List all profile names."""
    return profile.names()


@profile_cmd_group.command()
@click.argument("name", required=False)
def use(name: str):
    """Switch to a specific profile. If no name specified, launch interactive list to choose profile."""

    if name:
        if profile.use(name) is None:
            panic("No such profile: " + name)
    else:
        current = profile.name()
        choices = profile.names()
        if not choices:
            panic("No profile available.")

        ret = questionary.select("Select profile", choices, default=current, show_selected=True).ask()
        if ret:
            if profile.use(ret) is None:
                panic("No such profile: " + name)
        else:
            # aborted
            return "", 1


@profile_cmd_group.command()
@click.option("--from-name", "-f", required=False)
@click.option("--to-name", "-t", required=True)
def copy(from_name: str, to_name: str):
    """Copy profile."""

    if not from_name:
        from_name = profile.name()

    data = profile.get(from_name)
    if not data:
        panic("No such profile: " + from_name)
    if profile.exists(to_name):
        panic("Profile already exists: " + to_name)
    profile.create(to_name, data, True)


@profile_cmd_group.command()
@click.argument("name", required=False)
def get(name: str):
    """Get content of a specific profile."""
    if name:
        data = profile.get(name)
        if data is None:
            panic(
                "Profile not found. Use 'hcs profile list' to show available profiles, or 'hcs profile init' to create one."
            )
    else:
        data = profile.current()
        if data is None:
            panic(
                "Default profile not set. Use 'hcs profile list' to show available profiles, 'hcs profile user <name>' to switch to a profile, or 'hcs profile init' to create one."
            )

    return data


@profile_cmd_group.command()
@click.argument("name")
def delete(name: str):
    """Delete a profile by name."""
    profile.delete(name)


@profile_cmd_group.command()
@click.argument("name", required=False)
def file(name: str):
    """Show file location of a profile by name."""
    if not name:
        name = profile.name()
    return profile.file(name)


@profile_cmd_group.command()
@click.argument("name", type=str)
@click.option(
    "--file",
    "-f",
    type=click.File("rt"),
    default=sys.stdin,
    help="Specify the template file name. If not specified, STDIN will be used.",
)
def create(name: str, file):
    """Create a profile from file or STDIN."""
    with file:
        text = file.read()
    try:
        data = json.loads(text)
    except Exception as e:
        panic(f"Invalid json {e}")

    profile.create(name, data)


@profile_cmd_group.command()
def name():
    """Get current profile name."""
    return profile.name()


@profile_cmd_group.command()
@click.argument("name", type=str, required=False)
def edit(name: str):
    """Launch platform-specific editor to edit the profile file."""

    if not name:
        name = profile.name()
    file_name = profile.file(name)
    util.launch_text_editor(file_name)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from urllib.parse import urlencode
from typing import Any, Callable


def _remove_none(obj: dict):
    keys = list(obj.keys())
    for k in keys:
        v = obj[k]
        if v is None:
            del obj[k]
    return obj


def with_query(url: str, **kwargs: Any) -> str:
    qs = urlencode(_remove_none(dict(kwargs)))
    if qs:
        if url.find("?") < 0:
            url += "?"
        else:
            url += "&"
        url += qs
    return url


class PageRequest:
    def __init__(self, fn_get_page: Callable, fn_filter: Callable = None, **kwargs):
        limit = kwargs.get("limit")
        self.limit = int(limit) if limit else 10
        self.fn_get_page = fn_get_page
        self.fn_filter = fn_filter
        self.query = _remove_none(dict(kwargs))
        if int(self.query.get("size", 0)) < 1:
            self.query["size"] = 20

    def get(self) -> list:
        ret = []
        page_index = 0

        while True:
            self.query["page"] = page_index

            query_string = urlencode(self.query)
            page = self.fn_get_page(query_string)
            if not page or not page.content:
                break

            if self.fn_filter:
                content = list(filter(self.fn_filter, page.content))
            else:
                content = page.content
            ret += content
            if len(ret) > self.limit:
                ret = ret[: self.limit]
                break
            if len(page.content) < self.query["size"]:
                break  # no more items
            page_index += 1

        return ret



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from copy import deepcopy
import time
import threading
from graphlib import TopologicalSorter
from graphviz import Digraph
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Callable
from .helper import PlanException
from hcs_core.ctxp import data_util
from hcs_core.util import exit


class walker:
    next = "next"
    stop = "stop"


class _NodeType:
    var = "var"
    env = "env"
    const = "const"
    provider = "provider"
    resource = "resource"
    runtime = "runtime"


class DAG:
    graph: dict[str, set[str]]
    data: dict[str, Any]

    def __init__(self):
        self.graph = {}
        self.data = {}
        self.extra_source = []

    def add(self, id: str, type: str, data: Any, dependencies=None):
        if id in self.data:
            raise PlanException("Node already added to graph: " + id)
        self.data[id] = {"type": type, "data": data, "ghost": False}
        self.graph[id] = set(dependencies) if dependencies else set()

    def validate(self, extra):
        all_keys = self.data.keys()
        for k, v in self.graph.items():
            for d in v:
                if d in all_keys:
                    continue
                if d in extra:
                    continue
                raise PlanException(f"Blueprint error: target dpendency not found: from={k}, target={d}")


def process_blueprint(
    blueprint,
    fn_process_node: Callable,
    continue_on_error: bool = False,
    reverse: bool = False,
    concurrency: int = 3,
    target_node_name: str = None,
    include_dependencies: bool = True,
):
    dag = _build_graph(blueprint, reverse, target_node_name, include_dependencies)

    def fn_process_node_impl(name: str):
        if name.startswith("priority/"):
            actual_name = name[len("priority/") :]
            return fn_process_node(actual_name)
        if dag.data[name]["ghost"]:
            return walker.next
        return fn_process_node(name)

    return _walkthrough(dag, fn_process_node_impl, continue_on_error, concurrency)


def _filter_dag_by_target_node(dag: DAG, target_node_name, include_dependencies):
    if include_dependencies:
        # Remove all nodes that are not dependencies (both direct and indirect) of the target
        all_nodes = set(dag.graph.keys())
        dependencies = set()
        dependencies.add(target_node_name)
        open_nodes = set(dag.graph[target_node_name])
        while open_nodes:
            n = open_nodes.pop()
            if n in dependencies:
                continue
            dependencies.add(n)
            open_nodes |= dag.graph[n]

        to_delete = all_nodes - dependencies
        for n in to_delete:
            del dag.graph[n]
            del dag.data[n]
    else:
        # keep only the target node
        dag.data = {target_node_name: dag.data[target_node_name]}
        dag.graph = {target_node_name: set()}
    pass


def _build_graph(blueprint, reverse: bool, target_node_name: str, include_dependencies: bool):
    dag = DAG()
    extra = blueprint.keys()

    def add_node(name, type, obj):
        dependencies = set()

        # identify dependencies due to variables
        variables = data_util.deep_find_variables(obj)
        for v in variables:
            i = v.find(".")
            resource_name = v if i < 0 else v[:i]
            if resource_name == "var":
                continue  # ignore dependencies to var, which is a common thing.
            dependencies.add(resource_name)

        # identify dependencies due to explicit "after" statement
        after = obj.get("after")
        if after:

            def _add(t):
                if t in dependencies:
                    raise PlanException(
                        f"Invalid blueprint: statement after contains a dependency that is already implicitly created. This is not necessary. Resource: {name}, unexpected after-key: {t}"
                    )
                dependencies.add(t)

            if isinstance(after, str):
                _add(after)
            else:
                for v in after:
                    _add(v)

        # identify dependencies due to provider
        if type == _NodeType.resource:
            provider_type = _get_provider_id(obj)
            dependencies.add(provider_type)
        dag.add(name, type, obj, dependencies)

    required_providers = set()
    for k, v in blueprint["runtime"].items():
        add_node(k, _NodeType.runtime, v)
    for k, v in blueprint["resource"].items():
        required_providers.add(_get_provider_id(v))
        add_node(k, _NodeType.resource, v)
    default = blueprint["default"]
    if default:
        add_node("default", _NodeType.const, default)
    var = blueprint["var"]
    if var:
        add_node("var", _NodeType.var, var)
    providers = blueprint["provider"]
    if providers:
        for provider_id, data in providers.items():
            add_node(provider_id, _NodeType.provider, data)
    # Provider config is optional. Add place-holder nodes for all missing providers.
    for provider_id in required_providers:
        if provider_id not in dag.data:
            add_node(provider_id, _NodeType.provider, {})

    dag.validate(extra)

    if reverse:
        _reverse_dag(dag)
        _handle_priority_override(blueprint, dag)

    if target_node_name:
        _filter_dag_by_target_node(dag, target_node_name, include_dependencies)

    return dag


def _handle_priority_override(blueprint, dag):
    priority_list = []
    for k, v in blueprint["runtime"].items():
        p = v.get("destroyPriority")
        if p is not None:
            priority_list.append({"priority": p, "name": k, "data": v})

    if not priority_list:
        return

    list.sort(priority_list, key=lambda i: i["priority"])

    leaves = []
    for k, v in dag.graph.items():
        if not v:
            leaves.append(k)

    last_priority = None
    for i in priority_list:
        dependencies = [last_priority] if last_priority else []
        name = "priority/" + i["name"]
        dag.add(name, _NodeType.runtime, i["data"], dependencies)
        last_priority = name
        dag.data[i["name"]]["ghost"] = True

    for leaf in leaves:
        dag.graph[leaf].add(last_priority)


def _get_provider_id(node_data):
    kind = node_data["kind"]
    provider_type = kind[: kind.find("/")]
    if provider_type == "runtime":
        raise PlanException("This is a regression bug. Must not be runtime here")
    return provider_type


def _walkthrough(dag: DAG, fn_process_node: Callable, continue_on_error: bool, concurrency: int):
    topological_sorter = TopologicalSorter(dag.graph)
    topological_sorter.prepare()
    lock = threading.Lock()

    flag_stop = threading.Event()

    flags = {"err": None}

    def process_node(node_id):
        try:
            if flag_stop.is_set() or exit.is_set():
                return
            ret = fn_process_node(node_id)
            with lock:
                if ret == walker.stop:
                    flag_stop.set()
                topological_sorter.done(node_id)
        except KeyboardInterrupt as e:
            with lock:
                flags["err"] = e
            exit.set()
        except Exception as e:
            with lock:
                flags["err"] = e
                if continue_on_error:
                    topological_sorter.done(node_id)
                else:
                    flag_stop.set()
        except SystemExit as e:
            with lock:
                flags["err"] = e
            exit.set()

    # with ThreadPoolExecutor(max_workers=concurrency, thread_name_prefix="dag-walker") as executor:

    #     def close():
    #         flag_stop.set()
    #         executor.shutdown(wait=False, cancel_futures=True)

    #     atexit.register(close)

    #     while topological_sorter.is_active():
    #         if flag_stop.is_set():
    #             break
    #         with lock:
    #             if not continue_on_error and flags["err"]:
    #                 break
    #             read_nodes = topological_sorter.get_ready()
    #         if not len(read_nodes):
    #             time.sleep(0.1)
    #             continue
    #         for node in read_nodes:
    #             executor.submit(process_node, node)
    #     executor.shutdown(wait=True)
    #     if flags["err"]:
    #         raise flags["err"]

    #     atexit.unregister(close)

    try:
        executor = ThreadPoolExecutor(max_workers=concurrency, thread_name_prefix="dag-walker")

        while topological_sorter.is_active():
            if flag_stop.is_set():
                break
            with lock:
                if not continue_on_error and flags["err"]:
                    break
                read_nodes = topological_sorter.get_ready()
            if not len(read_nodes):
                time.sleep(0.1)
                continue
            for node in read_nodes:
                executor.submit(process_node, node)
    except KeyboardInterrupt as e:
        exit.set()
        raise e
    except Exception as e:
        flag_stop.set()
        raise e
    finally:
        executor.shutdown(wait=False)
        ex = flags["err"]
        if ex:
            raise ex  # pylint: disable=raising-bad-type


def _has_indirect_dependency(tree: dict, from_node: str, to_node: str):
    deps = list(tree[from_node])
    deps.remove(to_node)
    while deps:
        n = deps.pop()
        new_deps = tree[n]
        if to_node in new_deps:
            return True
        deps += new_deps
    return False


def graph(blueprint: dict, state: dict, reverse: bool, target_resource: str, include_dependencies: bool) -> Digraph:
    dag = _build_graph(blueprint, reverse, target_resource, include_dependencies)
    topological_sorter = TopologicalSorter(dag.graph)
    sorted_nodes = list(topological_sorter.static_order())
    graph = Digraph(name=f"Deployment {blueprint['deploymentId']}", comment="Simplified")

    class styles:
        deployed = {"style": "filled", "fillcolor": "lightgrey"}

    # https://graphviz.org/doc/info/shapes.html#:~:text=There%20are%20three%20main%20types,and%20an%20HTML%2Dlike%20label.
    node_type_style_map = {
        _NodeType.const: {"shape": "tab"},
        _NodeType.provider: {"shape": "component"},
        _NodeType.resource: {
            # empty
        },
        _NodeType.var: {"shape": "note"},
        _NodeType.runtime: {"shape": "hexagon"},
    }
    GHOST = {"style": "dashed"}

    def _is_deployed(node_name):
        data = state["output"].get(node_name)
        if isinstance(data, list):
            return any(d for d in data)
        return data

    def _get_node_style(node_name):
        n = dag.data[node_name]
        attr = {} | node_type_style_map[n["type"]]
        if n["ghost"]:
            attr |= GHOST

        if _is_deployed(node_name):
            attr |= styles.deployed
        return attr

    for node in sorted_nodes:
        attrs = _get_node_style(node)
        graph.node(node, **attrs)

    # Add the edges based on the dependencies
    for node, dependencies in dag.graph.items():
        for dependency in dependencies:
            if _has_indirect_dependency(dag.graph, from_node=node, to_node=dependency):
                continue  # simplify the diagram by removing ommitable
            graph.edge(dependency, node)

    return graph


def _reverse_dag(dag: DAG):
    old_dep = deepcopy(dag.graph)
    dag.graph = {}
    for k, h in old_dep.items():
        dag.graph[k] = set()

    def add_dep(from_node: str, to_node: str):
        holder = dag.graph[from_node]
        holder.add(to_node)

    for k, h in old_dep.items():
        for t in h:
            if t in old_dep:
                add_dep(t, k)
            else:
                # This is an external dependency and not part of our node
                pass


def _test(reverse):
    print("reverse=", reverse)
    dag = DAG()
    dag.add("a", "a", None)
    dag.add("b1", "b1", {"a"})
    dag.add("b2", "b2", {"a"})
    dag.add("c", "c", {"b1"})
    dag.add("d", "d", {"b2", "c"})

    if reverse:
        _reverse_dag(dag)

    def fn_proc(id):
        print("-> ", id)
        print("sleeping", id)
        time.sleep(2)
        print("<- ", id)

    _walkthrough(dag, fn_proc, False, 3)
    print("exit")


if __name__ == "__main__":
    _test(False)
    _test(True)



import ssl
import hashlib
import socket
from urllib.parse import urlparse


def get_cert_thumbprints(url, verify_cert: bool = False):
    # Parse the URL to extract the hostname and port
    parsed_url = urlparse(url)
    hostname = parsed_url.hostname
    port = parsed_url.port or 443  # Use port 443 if no port is specified

    # Create a context with default settings and system's certificates
    context = ssl.create_default_context()
    if not verify_cert:
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE

    # Connect to the server
    with socket.create_connection((hostname, port)) as sock:
        with context.wrap_socket(sock, server_hostname=hostname) as sslsock:
            # Get the DER-encoded certificate
            der_cert = sslsock.getpeercert(binary_form=True)
            # Compute the SHA256 and SHA1 fingerprints
            sha256_fingerprint = hashlib.sha256(der_cert).hexdigest().upper()
            sha1_fingerprint = hashlib.sha1(der_cert).hexdigest().upper()
            return {"SHA256": sha256_fingerprint, "SHA1": sha1_fingerprint}


if __name__ == "__main__":
    url = "https://pod30-vc1.pod30.hcs.steslabs.net"
    ret = get_cert_thumbprints(url)
    print(ret)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import httpx
import json
import logging

log = logging.getLogger(__name__)

log_http_details = False


def _raise_on_4xx_5xx(response: httpx.Response):
    if not response.is_success:
        response.read()
        if len(response.text) > 0:
            text = _try_formatting_json(response.text)
            log.debug(text)

    response.raise_for_status()


def _try_formatting_json(text: str):
    try:
        return json.dumps(json.loads(text), indent=4)
    except Exception:
        return text


def _log_request(request):
    if not log.isEnabledFor(logging.DEBUG):
        return
    log.debug("\n")
    log.debug(f"--> {request.method} {request.url}")
    log.debug(f"--> {request.headers}")

    if len(request.content) > 0:
        text = _try_formatting_json(request.content)
        log.debug(text)


def _log_response(response: httpx.Response):
    if not log.isEnabledFor(logging.DEBUG):
        return
    request = response.request
    log.debug(f"<-- {request.method} {request.url} - {response.status_code}")
    log.debug(f"<-- {request.headers}")
    response.read()
    if len(response.text) > 0:
        text = _try_formatting_json(response.text)
        log.debug(text)
    log.debug("\n")


class CspClient:
    def __init__(self, url: str, oauth_token: dict = None, org_id: str = None) -> None:
        self._base_url = url
        self._oauth_token = oauth_token
        self._org_id = org_id

        self._client = httpx.Client(
            base_url=url,
            timeout=20,
            event_hooks={
                "request": [_log_request],
                "response": [_log_response, _raise_on_4xx_5xx],
            },
        )

    def login_with_api_token(self, api_token: str) -> dict:
        # https://console-stg.cloud.vmware.com/csp/gateway/authn/api/swagger-ui.html#/Authentication/getAccessTokenByApiRefreshTokenUsingPOST

        # curl -X 'POST' \
        # 'https://console.cloud.vmware.com/csp/gateway/am/api/auth/api-tokens/authorize' \
        # -H 'accept: application/json' \
        # -H 'Content-Type: application/x-www-form-urlencoded' \
        # -d 'refresh_token=<the-refresh-token>'

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
        # <no org id for this API>
        resp = self._client.post(
            "/csp/gateway/am/api/auth/api-tokens/authorize", headers=headers, data=f"api_token={api_token}"
        )
        self._oauth_token = resp.json()
        return self._oauth_token

    def login_with_client_id_and_secret(self, client_id: str, client_secret: str, org_id: str) -> dict:
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
        params = {
            "grant_type": "client_credentials",
        }
        if org_id:
            params["orgId"] = org_id
        resp = self._client.post(
            "/csp/gateway/am/api/auth/authorize",
            auth=(client_id, client_secret),
            headers=headers,
            params=params,
        )
        self._oauth_token = resp.json()
        return self._oauth_token

    # def get_oauth_token(self, force=False):
    #     if self._oauth_token and not force:
    #         return self._oauth_token

    #     if self._api_token:
    #         resp = self._login_by_api_token()
    #     elif self._refresh_token:
    #         if self._grant_type == 'refresh_token':
    #             resp = self._login_by_refresh_token_from_user_login()
    #         else:
    #             resp = self._login_by_refresh_token_from_api_token()
    #     elif self._refresh_token_from_api_token:
    #         resp = self._login_by_refresh_token_from_api_token()
    #     else:
    #         resp = self._login_by_client_id()
    #     try:
    #         oauth_token = resp.json()
    #     except:
    #         log.error(resp.content)
    #         raise
    #     self.use_oauth_token(oauth_token)
    #     return oauth_token

    # def use_oauth_token(self, oauth_token):
    #     access_token = oauth_token["access_token"]
    #     self._client.headers["authorization"] = "Bearer " + access_token

    #     token_ttl_seconds = int(oauth_token["expires_in"])
    #     self._token_expires_at = int(time.time() + token_ttl_seconds)

    #     decoded = jwt.decode(access_token, options={"verify_signature": False})
    #     log.debug(decoded)

    #     self._org_id = decoded["context_name"]
    #     self._decoded_jwt = decoded
    #     self._oauth_token = oauth_token

    # def _login_by_client_id(self):
    #     headers = {
    #         "Content-Type": "application/x-www-form-urlencoded",
    #         "Accept": "application/json",
    #     }
    #     params = {
    #         "grant_type": "client_credentials",
    #     }
    #     if self._org_id:
    #         params['orgId'] = self._org_id
    #     return self._client.post(
    #         "/csp/gateway/am/api/auth/authorize",
    #         auth=(self._client_id, self._client_key),
    #         headers=headers,
    #         params=params,
    #     )

    # def _login_by_refresh_token_from_api_token(self):
    #     #https://console-stg.cloud.vmware.com/csp/gateway/authn/api/swagger-ui.html#/Authentication/getAccessTokenByApiRefreshTokenUsingPOST

    #     # curl -X 'POST' \
    #     # 'https://console.cloud.vmware.com/csp/gateway/am/api/auth/api-tokens/authorize' \
    #     # -H 'accept: application/json' \
    #     # -H 'Content-Type: application/x-www-form-urlencoded' \
    #     # -d 'refresh_token=<the-refresh-token>'

    #     headers = {
    #         "Content-Type": "application/x-www-form-urlencoded",
    #         "Accept": "application/json",
    #     }
    #     # <no org id for this API>
    #     return self._client.post(
    #         "/csp/gateway/am/api/auth/api-tokens/authorize",
    #         headers=headers,
    #         data=f"api_token={self._refresh_token}"
    #     )
    # def _login_by_refresh_token_from_user_login(self):
    #     headers = {
    #         "Content-Type": "application/x-www-form-urlencoded",
    #         "Accept": "application/json",
    #     }
    #     # <no org id for this API>
    #     return self._client.post(
    #         "/csp/gateway/am/api/auth/tokens",
    #         headers=headers,
    #         data=f"grant_type=refresh_token&refresh_token={self._refresh_token}"
    #     )
    # def _login_by_api_token(self):
    #     # curl -X 'POST' \\n  'https://console-stg.cloud.vmware.com/csp/gateway/am/api/auth/api-tokens/authorize' \\n  -H 'accept: application/json' \\n  -H 'Content-Type: application/x-www-form-urlencoded' \\n  -d 'api_token=vA24tXLuWUlDmu-84ar5nl2zItvctRBKPOUyuBnxxPfOiRjEef1jJmDSW_IxRDMP'
    #     headers = {
    #         "Content-Type": "application/x-www-form-urlencoded",
    #         "Accept": "application/json",
    #     }
    #     params = {
    #         "grant_type": "client_credentials",
    #     }
    #     if self._org_id:
    #         params['orgId'] = self._org_id
    #     return self._client.post(
    #         "/csp/gateway/am/api/auth/api-tokens/authorize",
    #         headers=headers,
    #         params=params,
    #         data=f"api_token={self._api_token}",
    #     )

    def get_org_details(self, org_id: str):
        resp = self._client.get(f"/csp/gateway/am/api/orgs/{org_id}")
        return resp.json()



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import re
import subprocess
import json
import yaml
import types
import sys
import httpx
import traceback
import click
from typing import Any, Callable
import questionary


class CtxpException(Exception):
    pass


def error(reason: any, return_code: int = 1) -> tuple[CtxpException, int]:
    # Shortcut if the reason is an Exception already
    if isinstance(reason, Exception):
        return reason, return_code

    # Convert reason to string and wrap as CtxpException
    if isinstance(reason, str):
        pass
    elif isinstance(reason, dict):
        reason = json.dumps(reason, indent=4)
    else:
        reason = str(reason)
    return CtxpException(reason), return_code


def validate_error_return(reason, return_code):
    if return_code == 0:
        raise CtxpException("Invalid return code. return_code must not be 0 (success) in error situation.")
    if not isinstance(return_code, int):
        raise CtxpException("Invalid return code. return_code must be integer, but got " + type(return_code).__name__)


def print_output(data: Any, args: dict, file=sys.stdout):
    output = args.get("output", "json")
    fields = args.get("field")
    id_only = args.get("id_only", False)
    first = args.get("first", False)

    if type(data) is str:
        text = data
    elif isinstance(data, Exception):
        text = f"{type(data).__name__}: {data}"
    else:
        try:
            data = _convert_generator(data)
            if first and isinstance(data, list):
                data = data[0]

            if id_only:
                if fields:
                    raise CtxpException("--id-only and --fields should not be used together.")
                data = _convert_to_id_only(data)
            elif fields:
                data = _filter_fields(data, fields)

            if output is None or output == "json":
                text = json.dumps(data, default=vars, indent=4)
            elif output == "json-compact":
                text = json.dumps(data, default=vars)
            elif output == "yaml":
                from . import jsondot

                text = yaml.dump(jsondot.plain(data), sort_keys=False)
            elif output == "text":
                if isinstance(data, list):
                    text = ""
                    for i in data:
                        line = i if type(i) is str else json.dumps(i)
                        text += line + "\n"
                elif isinstance(data, dict):
                    text = json.dumps(data, indent=4)
                elif isinstance(data, str):
                    text = data
                else:
                    text = json.dumps(data, indent=4)
            elif output == "table":
                formatter = args["format"]
                text = formatter(data)
            else:
                raise Exception(f"Unexpected output format: {output}")
        except Exception as e:
            traceback.print_exc()
            print("Fail converting object:", e, file=sys.stderr)
            text = data
    print(text, end="", file=file, flush=True)


def print_error(error):
    critical_errors = [KeyError, TypeError]
    for ex in critical_errors:
        if isinstance(error, ex):
            traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)
            break
    msg = error_details(error)
    print(msg, file=sys.stderr, flush=True)


def _convert_generator(data: Any):
    if isinstance(data, types.GeneratorType):
        return list(data)
    return data


def _convert_to_id_only(data: Any):
    def _get_id(o):
        return o.get("id") if isinstance(o, dict) else o

    if isinstance(data, list):
        ret = []
        for d in data:
            ret.append(_get_id(d))
        return ret

    if isinstance(data, dict):
        return _get_id(data)

    return data


def _filter_fields(obj: Any, fields: str):
    parts = fields.split(",")

    def _filter_obj(o):
        if not isinstance(o, dict):
            return o
        for k in list(o.keys()):
            if k not in parts:
                del o[k]
        return o

    if isinstance(obj, list):
        return list(map(_filter_obj, obj))
    return _filter_obj(obj)


def panic(reason: Any = None, code: int = 1):
    if isinstance(reason, Exception):
        text = error_details(reason)
    else:
        text = str(reason)
    print(text, file=sys.stderr, flush=True)
    os._exit(code)


def launch_text_editor(file_name: str, default_editor: str = None):
    editor = os.environ.get("EDITOR", default_editor)
    if not editor:
        if os.name == "nt":
            os.system(file_name)
            return
        else:
            editor = "vi"

    cmd = editor + " " + file_name
    subprocess.call(cmd, shell=True)


def choose(prompt: str, items: list, fn_get_text: Callable = None, selected=None, select_by_default=True):
    if len(items) == 0:
        panic(prompt + " ERROR: no item available.")

    if fn_get_text is None:
        fn_get_text = lambda t: str(t)

    if select_by_default and len(items) == 1:
        ret = items[0]
        print(prompt + " only one item available. Select by default: " + fn_get_text(ret))
        return ret

    choices = []
    size = len(items)
    for i in range(size):
        label = fn_get_text(items[i])
        # if label in choices:
        #    raise Exception("Problem with the provided fn_get_text: generates non-unique label. Item=" + label)
        choices.append(label)

    # hack workaround bug of the questionary lib
    selected_key = fn_get_text(selected) if selected else None
    k = questionary.select(prompt, choices, default=selected_key, show_selected=True).ask()
    if k is None:
        panic()
    for i in range(size):
        if k == choices[i]:
            return items[i]
    raise Exception("This is a bug and should not happen")


def input_array(prompt: str, default: list[str] = None):
    default_value = ",".join(default) if default else None

    input_value = click.prompt(prompt, default_value)
    if not input_value:
        return []
    parts = input_value.split(",")
    ret = []
    for p in parts:
        ret.append(p.strip())
    return ret


def error_details(e):
    if isinstance(e, Exception):
        if isinstance(e, CtxpException):
            details = str(e)
        else:
            details = e.__class__.__name__
            msg = str(e)
            if msg:
                details += ": " + msg
        cause = e.__cause__
        if cause and cause != e:
            details += " | Caused by: " + error_details(cause)

        if isinstance(e, httpx.HTTPStatusError):
            details += "\n" + e.response.text
        return details
    else:
        return str(e)


def avoid_trace_for_ctrl_c():
    import sys

    def my_except_hook(exctype, value, traceback):
        if exctype == KeyboardInterrupt:
            print("Aborted (KeyboardInterrupt).", flush=True)
            sys.exit(1)
        else:
            sys.__excepthook__(exctype, value, traceback)

    sys.excepthook = my_except_hook


def parse_kv_pairs(kv_pair_list) -> dict:
    if not kv_pair_list:
        return

    ret = {}

    def _parse_pair(kv_pair: str):
        parts = kv_pair.split("=")
        if len(parts) != 2:
            raise CtxpException(f"Invalid property pair format. Expect format 'key=value', got '{kv_pair}'.")
        return parts[0], parts[1]

    for pair in kv_pair_list:
        k, v = _parse_pair(pair)
        if k in ret:
            raise CtxpException("Invalid property parameter. Key already specified: " + k)
        ret[k] = v
    return ret


def flatten_dict(data, fields_mapping):
    flattened_data = []
    for item in data:
        flattened_item = {}
        for field_path, new_field_name in fields_mapping.items():
            current_value = item
            field_path_list = field_path.split(".")
            for field in field_path_list:
                if isinstance(current_value, dict) and field in current_value:
                    current_value = current_value[field]
                else:
                    current_value = None
                    break
            flattened_item[new_field_name] = current_value
        flattened_data.append(flattened_item)
    return flattened_data


def strip_ansi(text):
    # Regular expression to match ANSI escape sequences
    ansi_escape = re.compile(r"\x1b\[[0-9;]*m")
    return ansi_escape.sub("", text)


def format_table(data: list, fields_mapping: dict, columns_to_sum: list = None):
    from tabulate import tabulate

    flattened_data = flatten_dict(data, fields_mapping)
    try:
        headers = list(fields_mapping.values())
        table = [[item.get(field) for field in headers] for item in flattened_data]

        if columns_to_sum:
            columns_to_sum_indices = {col: headers.index(col) for col in columns_to_sum}
            footer = [""] * len(headers)
            footer[0] = "Total"
            for col_name, col_index in columns_to_sum_indices.items():
                total = 0
                for row in table:
                    v = row[col_index]
                    if isinstance(v, str):
                        v = strip_ansi(v)
                        v = int(v)
                    elif isinstance(v, int):
                        pass
                    else:
                        raise Exception(f"Unexpected cell value type. Type={type(v)}, value={v}")
                    total += v
                footer[col_index] = total
            separator = ["-" * len(header) for header in headers]
            table += [separator, footer]
    except Exception as e:
        traceback.print_exc()

    return tabulate(table, headers=headers) + "\n"



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from copy import deepcopy
import pathlib
import shutil
import os
from typing import Any
from .util import panic, CtxpException
from .jsondot import dotdict, dotify
from . import state
from .data_util import load_data_file, save_data_file

_repo_path: str = None
_active_profile_name: str = None
_data: dotdict = None
_auth_cache: dict = None


def _set_active_profile_name(name):
    global _active_profile_name
    _active_profile_name = name
    state.set("active_profile", name)


def init(repo_path: str, active_profile_name: str = None) -> None:
    if active_profile_name:
        _set_active_profile_name(active_profile_name)

    global _repo_path
    _repo_path = repo_path


def path(profile_name: str = None) -> str:
    """Get directory name of a profile"""
    if not profile_name:
        profile_name = name()
    return os.path.join(_repo_path, profile_name)


def create(name: str, data: dict, auto_use: bool = True, overwrite: bool = False):
    if not overwrite and exists(name):
        if auto_use:
            use(name)
        raise CtxpException("Profile already exists: " + name)
    file_path = file(name)
    pathlib.Path(os.path.dirname(file_path)).mkdir(parents=True, exist_ok=True)
    save_data_file(data, file(name), "yaml", 0o600)
    if auto_use:
        use(name)
    return get(name)


def copy(from_name: str, to_name: str, overwrite: bool = True):
    if exists(to_name) and not overwrite:
        raise CtxpException("Profile already exists: " + to_name)
    data = get(from_name)
    if not data:
        raise CtxpException("Profile does not exist: " + from_name)
    create(to_name, deepcopy(data), auto_use=False)


def current(reload: bool = False, exit_on_failure=True, exclude_secret: bool = False) -> dict:
    """Get content of the current active profile"""
    profile_name = name()
    data = get(profile_name, reload)

    if data is None and exit_on_failure:
        panic(
            "Profile not set. Use 'hcs profile use [profile-name]' to choose one, or use 'hcs profile init' to create default profiles."
        )

    if exclude_secret:
        data = dotdict(dict(data))
        csp_config = dict(data.get("csp", {}))
        if csp_config:
            csp_config["apiToken"] = None
            csp_config["clientSecret"] = None
            data["csp"] = csp_config
    return data


def save():
    """Save the current profile"""
    global _data
    if _data is None:
        return
    write(name(), _data)


def write(profile_name: str, data: dict):
    """Save profile by name"""
    save_data_file(data, file(profile_name), "yaml", 0o600)


def name() -> str:
    """Get the current active profile name"""
    global _active_profile_name
    if not _active_profile_name:
        _active_profile_name = state.get("active_profile", default="default")
        if not exists(_active_profile_name):
            _set_active_profile_name("default")
    return _active_profile_name


def use(name: str) -> str:
    """Use to the specified profile"""

    profile_exists = name in names()
    if not profile_exists:
        return

    _set_active_profile_name(name)
    return name


def names() -> list[str]:
    """List profile names"""
    ret = [f for f in os.listdir(_repo_path) if os.path.isdir(os.path.join(_repo_path, f))]
    ret.sort()
    return ret


def delete(profile_name: str = None) -> None:
    """Delete a profile"""

    if profile_name is None:
        profile_name = name()
    global _active_profile_name, _data
    if _active_profile_name == profile_name:
        _active_profile_name = "default"
        _data = None

    shutil.rmtree(os.path.join(_repo_path, profile_name), ignore_errors=True)


def get(profile_name: str, reload: bool = False, default=None) -> dotdict:
    """Get profile data by name"""
    if profile_name == name():
        global _data
        if _data is not None and not reload:
            return _data
        data = load_data_file(file(profile_name), default=default)
        if data is not None:
            data = dotify(data)
            _data = data
    else:
        data = load_data_file(file(profile_name), default=default)
        if data is not None:
            data = dotify(data)
    return data


def file(profile_name: str) -> str:
    """Get the file path of a profile"""
    return os.path.join(_repo_path, profile_name, "profile.yml")


def exists(profile_name: str) -> bool:
    if profile_name not in names():
        return False
    return os.path.isfile(file(profile_name))


class auth:
    @staticmethod
    def _file_name() -> str:
        return os.path.join(_repo_path, name(), ".auth")

    @staticmethod
    def get() -> dict:
        global _auth_cache
        if _auth_cache is None:
            _auth_cache = load_data_file(auth._file_name(), {}, "yaml")
        return dotify(deepcopy(_auth_cache))

    @staticmethod
    def set(data: dict) -> None:
        global _auth_cache
        import json

        if json.dumps(data) == json.dumps(_auth_cache):
            return
        _auth_cache = deepcopy(data)
        file_name = auth._file_name()
        save_data_file(data, file_name, "yaml", 0o600)

    @staticmethod
    def delete() -> None:
        file_name = auth._file_name()
        if os.path.exists(file_name):
            os.unlink(file_name)
        global _auth_cache
        _auth_cache = None


# --------------------------------------------------


def _nested_obj_to_plain_dict(obj: Any, path: str, result: dotdict) -> None:
    t = type(obj)
    if t is str or t is int or t is float or t is bool:
        result[path] = obj
    elif t is dict or t is dotdict:
        for k in obj:
            v = obj[k]
            _nested_obj_to_plain_dict(v, path + "." + k, result)
    elif t is list:
        for i in range(len(obj)):
            v = obj[i]
            _nested_obj_to_plain_dict(v, path + "[" + i + "]", result)
    else:
        raise Exception(f"TODO: type={t}")


def plain() -> dotdict:
    _plain = dotdict()
    data = current()
    for k in data:
        _nested_obj_to_plain_dict(data[k], k, _plain)

    return _plain


# --------------------------------------------------



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import threading
import typing
import inspect
import re
from subprocess import CalledProcessError
from copy import deepcopy
import hcs_core.ctxp.data_util as data_util
from . import dag
from . import context
from .helper import PlanException, process_template
from .actions import actions
from .kop import KOP
from importlib import import_module

import logging

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)


def _prepare_data(data: dict, additional_context: dict, target_resource_name: str):
    if additional_context:
        common_items = data_util.get_common_items(additional_context.keys(), data.keys())
        if common_items:
            raise PlanException("blueprint and context have conflict keys: " + str(common_items))
        data.update(additional_context)
    blueprint, pending = process_template(data)

    if (
        target_resource_name
        and target_resource_name not in blueprint["resource"]
        and target_resource_name not in blueprint["runtime"]
    ):
        raise PlanException("Target resource or runtime not found: " + target_resource_name)

    for k, v in pending.items():
        if v.startswith("default.") or v.startswith("var."):
            if not k.find(".conditions."):
                raise PlanException(
                    f"Invalid blueprint. Unresolved static references. Variable not found: {v}. Required by {k}"
                )
    deployment_id = blueprint["deploymentId"]
    state_file = deployment_id + ".state.yml"
    prev = data_util.load_data_file(state_file, default={})
    state = {"pending": pending}
    state.update(blueprint)
    if "provider" not in state:
        state["provider"] = {}
    if "runtime" not in state:
        state["runtime"] = {}
    state["output"] = prev.get("output", {})
    state["destroy_output"] = prev.get("destroy_output", {})
    state["log"] = prev.get("log", {})
    exec_log = state["log"]
    if "create" not in exec_log:
        exec_log["create"] = []
    if "delete" not in exec_log:
        exec_log["delete"] = []

    context.set("deploymentId", state["deploymentId"])

    # try solving more variables
    # for k, v in state['output'].items():
    #     if not v:
    #         continue
    #     if not _has_successful_deployment(state, k):
    #         continue
    #     _resolve_pending_keys(state, k)

    return blueprint, state, state_file


# def _has_successful_deployment(state, name):
#     for v in state['log']['create']:
#         if v['name'] != name:
#             continue
#         if v['action'] == 'success':
#             return True


def apply(
    data: dict,
    additional_context: dict = None,
    target_resource_name: str = None,
    include_dependencies: bool = True,
    concurrency: int = 4,
):
    blueprint, state, state_file = _prepare_data(data, additional_context, target_resource_name)
    state["log"]["create"] = []  # clear deploy log

    def deploy_resource_or_runtime(name, res_data):
        _deploy_res(name, res_data, state)
        data_util.save_data_file(state, state_file)

    def process_resource_node(name: str):
        if name in blueprint["resource"]:
            res_data = blueprint["resource"][name]
            return deploy_resource_or_runtime(name, res_data)
        elif name in blueprint["runtime"]:
            res_data = blueprint["runtime"][name]
            return deploy_resource_or_runtime(name, res_data)
        elif name in blueprint["provider"]:
            # ignore.
            # Provide init is always ensured before running the resource
            # and does not follow deploy/destroy sequenct.
            pass
        else:
            # default, var, etc.
            pass

    try:
        dag.process_blueprint(
            blueprint=blueprint,
            fn_process_node=process_resource_node,
            continue_on_error=False,
            reverse=False,
            concurrency=concurrency,
            target_node_name=target_resource_name,
            include_dependencies=include_dependencies,
        )
    except CalledProcessError as e:
        raise PlanException(str(e))
    finally:
        data_util.save_data_file(state, state_file)

    if target_resource_name:
        return state["output"][target_resource_name]


def _parse_statement_for(res_name, state) -> typing.Tuple[str, list]:
    # for: email in var.userEmails
    res = state["resource"][res_name]
    for_statement = res.get("for")
    if not for_statement:
        return None, None
    pattern = r"(.+?)\s+in\s+(.+)"
    matcher = re.search(pattern, for_statement)

    def _raise_error(reason):
        raise PlanException(f"Invalid for statement: {reason}. Resource={res_name}, statement={for_statement}")

    if not matcher:
        _raise_error("Invalid syntax")
    var_name = matcher.group(1)
    values_name = matcher.group(2)
    values = _get_value_by_path(state, values_name, f"resource.{res_name}.for")
    if not isinstance(values, list):
        reason = "The referencing value is not a list. Actual type=" + type(values).__name__
        _raise_error(reason)

    return var_name, values


def _get_value_by_path(state, var_name, required_by_attr_path):
    i = var_name.find(".")
    if i < 0:
        resource_name = var_name
    else:
        resource_name = var_name[:i]

    def _raise(e):
        msg = f"Plugin error: '{var_name}' does not exist in the output of resource '{resource_name}', which is required by '{required_by_attr_path}'"
        raise PlanException(msg) from e

    if resource_name in state["resource"]:
        try:
            return data_util.deep_get_attr(state, "output." + var_name)
        except Exception as e:
            _raise(e)
    if resource_name in state:
        try:
            return data_util.deep_get_attr(state, var_name)
        except Exception as e:
            _raise(e)


def _get_value_by_path2(state, var_name):
    i = var_name.find(".")
    if i < 0:
        resource_name = var_name
    else:
        resource_name = var_name[:i]

    if resource_name in state["resource"] or resource_name in state["runtime"] or resource_name in state["provider"]:
        try:
            return data_util.deep_get_attr(state, "output." + var_name), True
        except Exception as e:
            log.debug(e)
            return None, False

    if resource_name in state:
        try:
            return data_util.deep_get_attr(state, var_name), True
        except Exception as e:
            log.debug(e)
            return None, False
    return None, False


def _resolve_node_vars(node: dict, state: dict):
    def _get_value(path):
        return _get_value_by_path2(state, path)

    return data_util.process_variables(node, _get_value)


def _get_res_text(handler, res_state: dict) -> str:
    if res_state and hasattr(handler, "text"):
        fn_text = getattr(handler, "text")
        if callable(fn_text):
            return fn_text(res_state)
    if isinstance(res_state, dict):
        name = res_state.get("name")
        id = res_state.get("id")
        if name and id:
            return f"{name} ({id})"
        if name:
            return name
        return id


def _deploy_res(name, res, state):
    def fn_deploy1(handler, res_data: dict, res_state: dict, fn_set_state: typing.Callable, kop: KOP):
        if _is_runtime(res):
            kop.start(KOP.MODE.create, handler.eta("create", res_data, state))
            new_state = handler.process(res_data, deepcopy(state))
            if new_state:
                fn_set_state(new_state)
            return

        if not res_state:
            action = actions.create
        else:
            action = handler.decide(res_data, res_state)

        if action == actions.skip:
            kop.skip("Already deployed")
            return

        if action == actions.recreate:
            with KOP(state, res["kind"], name) as kop2:
                kop2.id(_get_res_text(handler, res_state))
                kop2.start(KOP.MODE.delete, handler.eta(actions.delete, res_data, res_state))
                handler.destroy(res_data, res_state, False)
            action = actions.create

        new_state = None
        if action == actions.create or action is None:
            kop.start(KOP.MODE.create, handler.eta(actions.create, res_data, res_state))
            if _has_save_state(handler.deploy):

                def _hook_set_state(data):
                    fn_set_state(data)
                    kop.id(_get_res_text(handler, data))

                new_state = handler.deploy(res_data, res_state, _hook_set_state)
            else:
                new_state = handler.deploy(res_data, res_state)
            kop.id(_get_res_text(handler, new_state))
        elif action == actions.update:
            kop.id(_get_res_text(handler, res_state))
            kop.start(KOP.MODE.update, handler.eta(actions.update, res_data, res_state))
            if _has_save_state(handler.update):
                new_state = handler.update(res_data, res_state, fn_set_state)
            else:
                new_state = handler.update(res_data, res_state)
        else:
            raise PlanException(
                f"Unknown action. This is a problem of the concrete plugin.decide function. Plugin={name}, action={action}"
            )

        if new_state:
            fn_set_state(new_state)

    _handle_resource(name, res, state, True, fn_deploy1)


def _is_runtime(res):
    return "impl" in res


def _has_save_state(fn):
    signature = inspect.signature(fn)
    args = list(signature.parameters.keys())
    if len(args) < 3:
        return False
    name = args[2]
    if name == "save_state":
        return True
    p = signature.parameters[args[2]]
    return p.annotation == typing.Callable


def _assert_all_vars_resolved(data, name):
    def fn_on_value(path, value):
        if isinstance(value, str) and value.find("${") >= 0:
            raise PlanException(f"Unresolved variable '{path}' for plugin '{name}'. Value={value}")
        return value

    data_util.deep_update_object_value(data, fn_on_value)


_providers = {}
_provider_lock = threading.Lock()


def _get_provider_module_name(provider_id: str):
    # Trick: make default HCS provider in the same module as services,
    # so for dev the update all happens in a self-contained module, for better blast radius control.
    if provider_id == "dev":
        return "hcs_core.plan.provider.dev"
    if provider_id == "hcs":
        return "hcs_cli.provider.hcs"
    return f"hcs_ext_{provider_id}.provider"


def _ensure_provider(provider_id: str, module_name: str, state: dict, require_prep: bool):
    # Ensure provider initialized
    with _provider_lock:
        holder = _providers.get(provider_id)

        if not holder:
            holder = {"instance": None, "prepared": False, "error": None}
            _providers[provider_id] = holder

            try:
                provider = import_module(module_name)
                holder["instance"] = provider

                # Get provider data
                meta = state["provider"].get(provider_id)

                _resolve_node_vars(meta, state)

                data = meta.get("data") if meta else None
                if data:
                    _assert_all_vars_resolved(data, provider_id)
                else:
                    data = {}
            except Exception as e:
                holder["error"] = e
                log.error(f"Fail loading provider {provider_id}: {e}")

        if holder["error"]:
            raise holder["error"]

        if require_prep and not holder["prepared"]:
            holder["prepared"] = True
            log.debug("[init] Provider: %s", provider_id)
            try:
                state["output"][provider_id] = provider.prepare(data)
            except BaseException as e:
                holder["error"] = e
                log.error(f"Fail initializing provider {provider_id}: {e}")
                raise e
            log.debug("[ok  ] Provider: %s", provider_id)


def _get_handler(name: str, res: dict, state: dict, require_prep: bool = True):
    kind = _get_kind(name, res)

    def _get_resource_handler(provider_path):
        provider_id, res_handler_type = provider_path.split("/")
        res_handler_type = res_handler_type.replace("-", "_")

        provider_module_name = _get_provider_module_name(provider_id)
        _ensure_provider(provider_id, provider_module_name, state, require_prep)
        handler_module_name = provider_module_name + "." + res_handler_type
        return import_module(handler_module_name)

    if kind == "runtime":
        impl_name = res["impl"]
        if impl_name.find("/") > 0:
            # provider-specific impl
            return _get_resource_handler(impl_name)
        else:
            return import_module(impl_name)
    else:
        return _get_resource_handler(kind)


def _get_kind(name: str, res: dict) -> str:
    if "kind" in res:
        return res["kind"]
    elif "impl" in res:
        return "runtime"
    else:
        raise PlanException("Invalid definition. Neither kind nor impl attribute found. Resource name: " + name)


def get_common_items(iter1, iter2):
    return set(iter1).intersection(set(iter2))


def _handle_resource(name, res, state, for_deploy: bool, fn_process: typing.Callable):
    kind = _get_kind(name, res)

    kop_mode = KOP.MODE.create if for_deploy else KOP.MODE.delete
    with KOP(state, kind, name, kop_mode) as kop:

        conditions = res.get("conditions")
        if conditions:
            conditions = deepcopy(conditions)
            ret = _resolve_node_vars(conditions, state)
            unsatisfied_condition_name = _get_unsatisfied_condition_name(conditions)
            if unsatisfied_condition_name:
                kop.skip("Condition not met: " + unsatisfied_condition_name)
                return

        # resolve var
        data = res.get("data", {})
        if data:
            data = deepcopy(data)
            ret = _resolve_node_vars(data, state)
            if for_deploy:
                if ret["pending"]:
                    msg = f"Fail resolving variables for resource '{name}'. Unresolvable variables: {ret['pending']}"
                    raise PlanException(msg)
        state["resource"][name] = dict(res)
        state["resource"][name]["data"] = data

        handler = _get_handler(name, res, state)

        def _handle_resource_1(resource_data, resource_state, fn_set_state, kop) -> bool:
            if _is_runtime(res):  # runtime has no refresh
                pass
            else:
                new_state = handler.refresh(resource_data, resource_state)
                fn_set_state(new_state)
                resource_state = new_state

            fn_process(handler, resource_data, resource_state, fn_set_state, kop)

        for_var_name, values = _parse_statement_for(name, state)
        if for_var_name:
            # group

            if for_var_name in data:
                raise PlanException(
                    f"Invalid blueprint: variable name defined in for-statement already exists in data declaration. Resource: {name}. Conflicting names: {for_var_name}"
                )
            kop.id("(group)")
            kop.start()
            size = len(values)
            # ensure output array placeholder
            output = state["output"].get(name)
            if not output:
                output = []
                state["output"][name] = output
            while len(output) < size:
                output.append(None)
            for i in range(size):
                v = values[i]
                with KOP(state, kind, name + f"#{i}", kop_mode) as kop_per_item:
                    kop_per_item.id(str(i))
                    resource_state = output[i]

                    def _fn_set_state(o):
                        output[i] = deepcopy(o)

                    resource_data = deepcopy(data)
                    resource_data[for_var_name] = v
                    if v is None:
                        kop_per_item.skip("No input data")
                    else:
                        _handle_resource_1(resource_data, resource_state, _fn_set_state, kop_per_item)
        else:
            # Single resource

            resource_state = state["output"].get(name)

            def _fn_set_state(o):
                state["output"][name] = deepcopy(o)

            resource_data = deepcopy(data)
            _handle_resource_1(resource_data, resource_state, _fn_set_state, kop)


def _get_unsatisfied_condition_name(conditions):
    if not conditions:
        return
    for condition_name, expr in conditions.items():
        if not expr:
            return condition_name
        if isinstance(expr, str):
            if expr.find("${") >= 0:  # still have unresolved variables
                return condition_name
        # expr could be an object. It's already "True". So skip.


def _destroy_res(name, res_node, state, force):
    def fn_destroy1(handler, res_data: dict, res_state: dict, fn_set_state: typing.Callable, kop: KOP):
        if not res_state:
            kop.skip("Not found")
            return
        kop.id(_get_res_text(handler, res_state))
        kop.start(KOP.MODE.delete, handler.eta(actions.delete, res_data, res_state))
        ret = handler.destroy(res_data, res_state, force)
        state["destroy_output"][name] = deepcopy(ret)

        if _is_runtime(res_node):
            # No set empty data for runtime. Runtime is special and normally the data needs to be referenced in the next run.
            pass
        else:
            fn_set_state(None)

    _handle_resource(name, res_node, state, False, fn_destroy1)


def destroy(
    data,
    force: bool,
    target_resource_name: str = None,
    include_dependencies: bool = True,
    concurrency: int = 4,
    additional_context: dict = None,
):
    blueprint, state, state_file = _prepare_data(data, additional_context, target_resource_name)
    state["log"]["delete"] = []  # clear destroy log

    def destroy_resource(node_name):
        # ignore functional nodes (default, provider)
        node = blueprint["resource"].get(node_name)
        if not node:
            node = blueprint["runtime"].get(node_name)

        if not node:
            return dag.walker.next

        _destroy_res(node_name, node, state, force)
        data_util.save_data_file(state, state_file)
        return dag.walker.next

    try:
        dag.process_blueprint(
            blueprint=blueprint,
            fn_process_node=destroy_resource,
            continue_on_error=force,
            reverse=True,
            concurrency=concurrency,
            target_node_name=target_resource_name,
            include_dependencies=include_dependencies,
        )
    except CalledProcessError as e:
        raise PlanException(str(e))
    finally:
        data_util.save_data_file(state, state_file)


def graph(
    data: dict,
    additional_context: dict = None,
    reverse: bool = False,
    target_resource: str = None,
    include_dependencies: bool = True,
):
    blueprint, state, state_file = _prepare_data(data, additional_context, None)
    g = dag.graph(blueprint, state, reverse, target_resource, include_dependencies)
    return g


def get_deployment_data(data: dict, additional_context: dict = None, resource_name: str = None):
    blueprint, state, state_file = _prepare_data(data, additional_context, None)
    input = {}
    output = {}

    for k in blueprint["resource"]:
        if resource_name and k != resource_name:
            continue
        res_state = state["output"].get(k)
        res_node = state["resource"].get(k)
        output[k] = res_state
        if res_state:
            handler = _get_handler(k, res_node, state, require_prep=False)
            if isinstance(res_state, dict):
                res_state["_display"] = _get_res_text(handler, res_state)

        input[k] = res_node.get("data")
        _resolve_node_vars(input[k], state)

    for k in blueprint["runtime"]:
        if resource_name and k != resource_name:
            continue
        output[k] = state["output"].get(k)
        v = state["runtime"].get(k)
        if v:
            v = v.get("data")
            _resolve_node_vars(v, state)
        input[k] = v
    return input, output


def clear(
    data: dict,
    resource_name: str,
    additional_context: dict = None,
):
    blueprint, state, state_file = _prepare_data(data, additional_context, None)
    state["output"][resource_name] = None
    data_util.save_data_file(state, state_file)



from datetime import timedelta


def display(d: timedelta, use_double_digits: bool = False) -> str:
    years = int(d.days / 365)
    remaining = d.days % 365
    weeks = int(remaining / 7)
    days = remaining % 7

    hours = int(d.seconds / 3600)
    remaining = d.seconds % 3600
    minutes = int(remaining / 60)
    seconds = remaining % 60

    if use_double_digits:
        if years:
            return f"{years:02}y{weeks:02}w"
        if weeks:
            return f"{weeks:02}w{days:02}d"
        if days:
            return f"{days:02}d{hours:02}h"
        if hours:
            return f"{hours:02}h{minutes:02}m"
        if minutes:
            return f"{minutes:02}m{seconds:02}s"
        return f"{seconds}s"
    else:
        if years:
            return f"{years}y{weeks}w"
        if weeks:
            return f"{weeks}w{days}d"
        if days:
            return f"{days}d{hours}h"
        if hours:
            return f"{hours}h{minutes}m"
        if minutes:
            return f"{minutes}m{seconds}s"
        return f"{seconds}s"



from .core import apply, destroy, graph, get_deployment_data, clear
from .helper import PlanException, PluginException
from .actions import actions
from . import context
from .kop import attach_job_view



import os
import subprocess
import zipfile

import time

from hcs_core.ctxp import jsondot


def _print_output(text, name):
    if len(text) == 0:
        print(name + ": <EMPTY>")
    else:
        print(name + ":")
        lines = text.splitlines()
        for line in lines:
            print(">> " + line)


def run(cmd, check=True):
    print("CMD: " + cmd)
    parts = cmd.split(" ")
    proc = subprocess.run(
        parts,
        shell=False,
        check=False,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
    )
    if check and proc.returncode != 0:
        _print_output(proc.stdout, "STDOUT")
        _print_output(proc.stderr, "STDERR")
        proc.check_returncode()
    return proc


def _run_and_capture_output(cmd):
    proc = subprocess.run(cmd, shell=True, check=False, text=True, capture_output=True, encoding="utf-8")
    if proc.returncode != 0:
        _print_output(proc.stdout, "STDOUT")
        _print_output(proc.stderr, "STDERR")
        proc.check_returncode()
    return proc.stdout.strip()


def run_govc_guest_script(script, args):
    print("Running guest script: " + script + " " + args)
    cmd = f'govc guest.start {script} "{args} >/tmp/script-out 2>/tmp/script-err"'
    output = _run_and_capture_output(cmd)

    # wait for the process to finish
    cmd = f"govc guest.ps -X -x -json -p {output}"
    output = _run_and_capture_output(cmd)
    ret = jsondot.parse(output)

    output = _run_and_capture_output("govc guest.download /tmp/script-out -")
    _print_output(output, "STDOUT")
    output = _run_and_capture_output("govc guest.download /tmp/script-err -")
    if len(output) > 0:
        _print_output(output, "STDERR")

    if ret.ProcessInfo[0].ExitCode != 0:
        raise Exception("Fail running script: " + script)


def waitForVmPowerOff(iPath):
    current_millis = lambda: int(round(time.time() * 1000))
    start_time = current_millis()
    while current_millis() - start_time < 5 * 60 * 1000:
        p = subprocess.run(
            "govc vm.info -vm.ipath=" + iPath,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            shell=True,
        )
        if str(p.stdout).find("poweredOff") > 0:
            return
        time.sleep(5)
    raise Exception("Timeout customization. Check log in deployer VM.")


def zipdir(src_dir, target_file):
    with zipfile.ZipFile(target_file, "w", zipfile.ZIP_DEFLATED) as file:
        for root, dirs, files in os.walk(src_dir):
            for name in files:
                filePath = os.path.join(root, name)
                arcname = filePath[len(src_dir) :]
                file.write(filePath, arcname)


def dos2unix(file):
    file = os.path.abspath(file)
    content = ""
    outsize = 0
    with open(file, "rb") as infile:
        content = infile.read()
    with open(file, "wb") as output:
        for line in content.splitlines():
            outsize += len(line) + 1
            output.write(line + b"\n")
    stripped_bytes = len(content) - outsize
    if stripped_bytes != 0:
        print("dos2unix: Stripped %s bytes. %s" % (stripped_bytes, file))



import os
import os.path
import subprocess

import logging

log = logging.getLogger(__name__)


def run_cmd(cmd, shell: bool = True):
    cmd_text = " ".join(cmd)
    log.info(f"Running command: {cmd_text}")
    result = {}
    result["command"] = cmd_text
    # proc = subprocess.run(cmd, shell=shell, check=False, capture_output=True, timeout=60, encoding="utf-8")
    proc = subprocess.run(cmd, shell=shell, check=False, timeout=60, **subprocess_args())
    result["stdout"] = proc.stdout.decode("utf-8")
    result["stderr"] = proc.stderr.decode("utf-8")
    result["returnCode"] = proc.returncode
    return result


# Create a set of arguments which make a ``subprocess.Popen`` (and
# variants) call work with or without Pyinstaller, ``--noconsole`` or
# not, on Windows and Linux. Typical use::
#
#   subprocess.call(['program_to_run', 'arg_1'], **subprocess_args())
#
# When calling ``check_output``::
#
#   subprocess.check_output(['program_to_run', 'arg_1'],
#                           **subprocess_args(False))
def subprocess_args(include_stdout=True):
    # The following is true only on Windows.
    if hasattr(subprocess, "STARTUPINFO"):
        # On Windows, subprocess calls will pop up a command window by default
        # when run from Pyinstaller with the ``--noconsole`` option. Avoid this
        # distraction.
        si = subprocess.STARTUPINFO()
        si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        # Windows doesn't search the path by default. Pass it an environment so
        # it will.
        env = os.environ
    else:
        si = None
        env = None

    # ``subprocess.check_output`` doesn't allow specifying ``stdout``::
    #
    #   Traceback (most recent call last):
    #     File "test_subprocess.py", line 58, in <module>
    #       **subprocess_args(stdout=None))
    #     File "C:\Python27\lib\subprocess.py", line 567, in check_output
    #       raise ValueError('stdout argument not allowed, it will be overridden.')
    #   ValueError: stdout argument not allowed, it will be overridden.
    #
    # So, add it only if it's needed.
    if include_stdout:
        ret = {"stdout": subprocess.PIPE}
    else:
        ret = {}

    # On Windows, running this from the binary produced by Pyinstaller
    # with the ``--noconsole`` option requires redirecting everything
    # (stdin, stdout, stderr) to avoid an OSError exception
    # "[Error 6] the handle is invalid."
    ret.update(
        {
            "stdin": subprocess.PIPE,
            "stderr": subprocess.PIPE,
            "startupinfo": si,
            "env": env,
        }
    )
    return ret


# A simple test routine. Compare this output when run by Python, Pyinstaller,
# and Pyinstaller ``--noconsole``.
if __name__ == "__main__":
    # Save the output from invoking Python to a text file. This is the only
    # option when running with ``--noconsole``.
    with open("out.txt", "w") as f:
        try:
            txt = subprocess.check_output(["python", "--help"], **subprocess_args(False))
            f.write(txt)
        except OSError as e:
            f.write("Failed: " + str(e))



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# nanw @ 2023
# PKCE with authlib: https://docs.authlib.org/en/latest/specs/rfc7636.html

import threading
import time
from typing import Callable
import webbrowser
import secrets
import logging
from hcs_core.util.scheduler import ThreadPoolScheduler
from authlib.integrations.httpx_client import OAuth2Client
from authlib.oauth2.rfc7636 import create_s256_code_challenge
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import urlparse, parse_qs

log = logging.getLogger(__name__)

_server_address = ("127.0.0.1", 10762)
_callback_url = "/hcs-cli/oauth/callback"

_public_client_ids = {
    "production": "ldjmWbBAUcSB1w3HSbzoYcdKoloYFqT2dWK",
    "staging": "BbwHlgWt0vFwPUZMJTb5IKltynXjDmb46IO",
}

_auth_code_event = threading.Event()
_state = None

_auth_success_html = """
<!DOCTYPE html>
<html>
<head>
    <meta charset="utf-8" />
    <title>Login successfully</title>
    <style>
        body {
            font-family: 'Segoe UI', Tahoma, Geneva, Verdana, sans-serif;
        }

        code {
            font-family: Consolas, 'Liberation Mono', Menlo, Courier, monospace;
            display: inline-block;
            background-color: rgb(242, 242, 242);
            padding: 12px 16px;
            margin: 8px 0px;
        }
    </style>
</head>
<body>
    <h3>You have successfully logged into VMware Horizon Cloud Service.</h3>
    <p>You can close this window, and return to the terminal.</p>
    <br/>
    <p><a href="https://github.com/euc-eng/hcs-cli">HCS CLI</a> is in beta phase. <a href="https://github.com/euc-eng/hcs-cli/blob/main/doc/hcs-cli-cheatsheet.md">Cheatsheet</a>:</p>
    <code>
    # To get the login details: <br/>
    hcs login -d <br/><br/>
    # To login programmatically: <br/>
    hcs login --api-token &lt;my-CSP-user-API-token&gt; <br/><br/>
    # Switch to a different profile:<br/>
    hcs profile use <br/><br/>
    </code>
</body>
</html>
"""


class OAuthCallbackHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        def _respond(code: int, reason: str, content_type: str = "text/plain"):
            self.send_response(code)
            self.send_header("Content-type", content_type)
            self.end_headers()
            self.wfile.write(reason.encode())
            self.wfile.flush()

        try:
            # Parse the query parameters from the callback URL
            parsed_url = urlparse(self.path)
            log.debug(parsed_url)
            if parsed_url.path == _callback_url or parsed_url.path == "/hcsadmin/index.html":
                query_components = parse_qs(parsed_url.query)
                codes = query_components.get("code", None)
                states = query_components.get("state", None)

                # Check if the 'code' parameter is present in the callback URL
                if not codes:
                    _respond(400, "Missing authorization code")
                    return
                if not states:
                    _respond(400, "Missing state in server response")
                    return
                state = states[0]
                if _state != state:
                    _respond(401, "Response state mismatch")
                    return

                code = codes[0]
                # Process the authorization code
                log.debug(f"auth code {code}")
                _auth_code_event.value = code

                # Send a response back to the client
                _respond(200, _auth_success_html, "text/html")
            else:
                _respond(404, "Not found")
        finally:
            # Terminate the server
            _auth_code_event.set()
            threading.Thread(target=lambda: self.server.shutdown()).start()

    def log_message(self, *args):
        log.debug(*args)


def run_server_async():
    def run_server():
        log.debug("oauth callback httpd - start")
        httpd = HTTPServer(_server_address, OAuthCallbackHandler)
        httpd.serve_forever()
        log.debug("oauth callback httpd - exit")

    threading.Thread(target=run_server, daemon=True).start()


def do_oauth2_pkce(csp_url: str, client_id: str, org_id: str):
    # OAuth2 server details
    # authorization_endpoint = csp_url + '/csp/gateway/am/api/auth/api-tokens/authorize'
    discovery_endpoint = csp_url + "/csp/gateway/discovery"
    token_endpoint = csp_url + "/csp/gateway/am/api/auth/token"

    redirect_uri = f"http://{_server_address[0]}:{_server_address[1]}{_callback_url}"

    # Create an OAuth2Session instance
    session = OAuth2Client(client_id, redirect_uri=redirect_uri)

    # Generate a code verifier and code challenge

    code_verifier = secrets.token_urlsafe(64)
    code_challenge = create_s256_code_challenge(code_verifier)
    # code_challenge = CodeChallenge().from_code_verifier(code_verifier)

    # Create authorization URL
    # https://console-stg.cloud.vmware.com/csp/gateway/authn/api/swagger-ui.html#/Discovery/getDiscoveryUsingGET
    global _state
    authorization_url, _state = session.create_authorization_url(
        discovery_endpoint, orgId=org_id, code_challenge=code_challenge, code_challenge_method="S256"
    )

    log.debug(authorization_url)

    # Redirect the user to the authorization URL
    webbrowser.open(authorization_url, new=0, autoraise=True)

    try:
        _auth_code_event.wait()
    except:
        log.info("Aborted")
        return
    # Once the user is redirected back to your app with an authorization code, exchange it for an access token
    authorization_code = _auth_code_event.value
    log.debug("authorization_code: %s", authorization_code)
    if not authorization_code:
        log.debug("Login failed")
        return

    token = session.fetch_token(
        token_endpoint,
        auth=(client_id, ""),  # auth method required by CSP API for PKCE flow
        authorization_response=redirect_uri + "?code=" + authorization_code,
        code_verifier=code_verifier,
    )

    log.debug("token: %s", token)
    return token


def identify_client_id(csp_url: str) -> str:
    if csp_url.find("console.cloud") >= 0:
        return _public_client_ids["production"]
    if csp_url.find("https://connect.omnissa.com") >= 0:
        return _public_client_ids["production"]

    if csp_url.find("console-stg.cloud") >= 0:
        return _public_client_ids["staging"]
    raise Exception("Unknonw CSP url: " + csp_url)


def login_via_browser(csp_url: str, org_id: str):
    client_id = identify_client_id(csp_url)
    run_server_async()
    return do_oauth2_pkce(csp_url, client_id, org_id)


def create_oauth_client(
    oauth_token: dict, csp_url: str, fn_on_new_oauth_token: Callable = None, fn_custom_refresh: Callable = None
):
    client_id = identify_client_id(csp_url)

    if not fn_custom_refresh and client_id:
        client = OAuth2Client(
            client_id=client_id,
            client_secret="",  # Required by CSP auth
            token=oauth_token,
            token_endpoint=csp_url + "/csp/gateway/am/api/auth/token",
            token_endpoint_auth_method="client_secret_basic",
            update_token=fn_on_new_oauth_token,
        )
        log.debug("create_oauth_client - no custom refresher")
    else:
        # This is the case for production today, no csp-oauth-app yet,
        # and auth with API token case.
        client = OAuth2Client(
            # client_id=client_id,
            # client_secret="",  # Required by CSP auth
            token=oauth_token,
            # token_endpoint=csp_url + "/csp/gateway/am/api/auth/token",
            # token_endpoint_auth_method="client_secret_basic",
            # update_token=fn_on_new_oauth_token,
        )
        # start our own token refresh timer. Refresh using provided method
        if oauth_token and fn_on_new_oauth_token:
            _register_custom_refresher(client, oauth_token, fn_on_new_oauth_token, fn_custom_refresh)
        else:
            # not even initialized.
            pass
    return client


_scheduler: ThreadPoolScheduler = None


def _register_custom_refresher(
    client: OAuth2Client, oauth_token, fn_on_new_oauth_token: Callable, fn_custom_refresh: Callable
):
    log.debug("_register_custom_refresher")
    global _scheduler
    if not _scheduler:
        _scheduler = ThreadPoolScheduler()

    def get_refresh_delay(token):
        delay_seconds = token["expires_at"] - time.time() - 30
        log.debug(f"delay_seconds {delay_seconds}")

        # The expiry normally is 4 hrs but effectively it expires every 30 minutes.
        safe_delay_seconds = 25 * 60
        if delay_seconds > safe_delay_seconds:
            delay_seconds = safe_delay_seconds
        return delay_seconds

    def refresh_impl():
        if client.is_closed:
            return
        log.debug("_register_custom_refresher#refesh_impl")
        new_token = fn_custom_refresh()
        log.debug(f"_register_custom_refresher#refesh_impl: token={new_token}")
        client.token = new_token
        fn_on_new_oauth_token(new_token)
        delay = get_refresh_delay(new_token)
        _scheduler.submit(refresh_impl, delay)

    delay = get_refresh_delay(oauth_token)
    _scheduler.submit(refresh_impl, delay)


def refresh_oauth_token(old_oauth_token: dict, csp_url: str):
    with create_oauth_client(old_oauth_token, csp_url) as client:
        token_endpoint = csp_url + "/csp/gateway/am/api/auth/token"
        log.debug("Refresh auth token...")
        new_token = client.refresh_token(token_endpoint)
        log.debug(f"New auth token: {new_token}")
        if not new_token:
            raise Exception("CSP auth refresh failed.")
        if "cspErrorCode" in new_token:
            raise Exception(f"CSP auth failed: {new_token.get('message')}")
    return new_token


# def _test():
#     import jwt
#     import json
#     csp_url = 'https://console-stg.cloud.vmware.com'
#     org_id = '06e09ab5-dc8d-413c-bd57-2c1e929a412c' #stg horizonv2-syntest
#     token = interactive_login_via_browser(csp_url, org_id)
#     decoded = jwt.decode(token['access_token'], options={"verify_signature": False})
#     print(json.dumps(decoded, indent=4))

# if __name__ == '__main__':
#     _test()



import httpx
import logging

from hcs_core.ctxp import jsondot

log = logging.getLogger(__name__)


def parse(resp: httpx.Response):
    if not resp.content:
        return None
    content_type = resp.headers["Content-Type"]
    if content_type.startswith("text"):
        return resp.text
    if len(resp.content) > 3:
        try:
            data = resp.json()
            if isinstance(data, list):
                return [jsondot.dotify(obj) for obj in data]
            elif isinstance(data, dict):
                return jsondot.dotify(data)
        except:
            log.info("--- Fail parsing json. Dump content ---")
            log.info(resp.content)
            raise
    else:
        return None


def on404ReturnNone(func):
    try:
        resp = func()
        return parse(resp)
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            return None
        raise



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Optional
import threading
from time import sleep, monotonic
import sys
import os
from rich.live import Live
from rich.progress import Progress, Task, ProgressBar, SpinnerColumn, TextColumn, TimeRemainingColumn, ProgressColumn
from rich.table import Column
from rich.text import Text
from rich.console import RenderableType
import hcs_core.util.duration as duration


def _shorten_package_name(package_name, max_length):
    if len(package_name) <= max_length:
        return package_name

    parts = package_name.split(".")
    shortened_parts = []

    for i, part in enumerate(parts):
        shortened_parts.append(part[0])

        if i == len(parts) - 1:
            # Last package name
            remaining_length = max_length - len(".".join(shortened_parts))
            if len(part) > remaining_length:
                shortened_parts[-1] = "..." + part[-remaining_length + 3 :]
            else:
                shortened_parts[-1] = part
        else:
            new_package_name = ".".join(shortened_parts)

            if len(new_package_name) + len(parts) - i - 1 > max_length:
                break

    return ".".join(shortened_parts)


class _MyPlainBarColumn(ProgressColumn):
    def __init__(self) -> None:
        super().__init__(table_column=Column(min_width=10))

    def render(self, task: Task) -> ProgressBar:
        """Gets a progress bar widget for a task."""
        return ProgressBar(
            total=max(0, task.total) if task.total is not None else None,
            completed=max(0, task.completed),
            width=10,
            pulse=False,
            animation_time=task.get_time(),
            style="bar.back",
            complete_style="bar.complete",
            finished_style="bar.finished",
            pulse_style="bar.pulse",
        )


class _MyTimeRemainingColumn(ProgressColumn):
    """Renders estimated time remaining."""

    # Only refresh twice a second to prevent jitter
    max_refresh = 0.5

    def __init__(
        self,
        compact: bool = False,
        elapsed_when_finished: bool = False,
        table_column: Optional[Column] = None,
    ):
        self.compact = compact
        self.elapsed_when_finished = elapsed_when_finished
        super().__init__(table_column=table_column)

    def render(self, task: "Task") -> Text:
        """Show time remaining."""
        if self.elapsed_when_finished and task.finished:
            task_time = task.finished_time
            style = "progress.download"
        else:
            task_time = task.time_remaining
            style = "progress.remaining"

        if task.total is None:
            return Text("", style=style)

        if task_time is None:
            return Text("--:--" if self.compact else "-:--:--", style=style)

        # Based on https://github.com/tqdm/tqdm/blob/master/tqdm/std.py
        minutes, seconds = divmod(int(task_time), 60)
        hours, minutes = divmod(minutes, 60)

        if self.compact and not hours:
            formatted = f"{minutes:02d}:{seconds:02d}"
        else:
            formatted = f"{hours:d}:{minutes:02d}:{seconds:02d}"

        return Text(formatted, style=style)


class _MySpinnerColumn(SpinnerColumn):
    def render(self, task: "Task") -> RenderableType:
        if task.finished:
            text = self.finished_text
        elif task.stop_time:
            text = " "  # stopped. E.g. error.
        else:
            text = self.spinner.render(task.get_time())
        return text


def _timeout_to_seconds(timeout: str, default: int = 60) -> int:
    t = duration.to_seconds(timeout)
    if t == 0 and default:
        t = default
    return t


class JobView:
    def __init__(self, auto_close: bool = True):
        self._map = {}
        self._todo = set()
        self._auto_close = auto_close
        self._closed = False
        self._view_started = threading.Event()
        self._view_exited = threading.Event()

        w, h = os.get_terminal_size()
        msg_width = w - 40 - 1 - 1 - 1 - 10 - 1 - 5 - 1 - 1
        self._job_ctl = Progress(
            TextColumn("{task.description}", table_column=Column(max_width=40, min_width=10)),
            # SpinnerColumn(table_column=Column(min_width=1)),
            _MySpinnerColumn(table_column=Column(min_width=1)),
            # BarColumn(pulse_style='white'),
            _MyPlainBarColumn(),
            # TextColumn("[white][progress.percentage][white]{task.percentage:>3.0f}%"),
            # TaskProgressColumn(show_speed=True),
            TimeRemainingColumn(
                compact=True, elapsed_when_finished=True, table_column=Column(style="white", min_width=5)
            ),
            TextColumn("{task.fields[details]}", table_column=Column(max_width=msg_width, no_wrap=True)),
            get_time=monotonic,
        )

    def add(self, id: str, name: str):
        name = _shorten_package_name(name, 30)
        self._map[id] = self._job_ctl.add_task(name, start=False, details="")
        self._todo.add(id)

    def remove(self, id: str):
        task_id = self._map[id]
        self._job_ctl.remove_task(task_id)
        del self._map[id]
        self._todo.discard(id)

    def refresh(self):
        for task_id in self._map.values():
            task = self._job_ctl._tasks[task_id]

            if task.started and not task.finished and not task.stop_time:
                completed = monotonic() - task.start_time
                if completed >= task.total:
                    completed = task.total - 1
                self._job_ctl.update(task_id, completed=completed)

    def update(self, id: str, details: str):
        task_id = self._map[id]
        if not details:
            details = ""
        self._job_ctl.update(task_id, details=details)

    def start(self, id: str, timeout: str) -> None:
        task_id = self._map[id]
        if timeout:
            total = _timeout_to_seconds(timeout)
            self._job_ctl.update(task_id, total=total)
        self._job_ctl.start_task(task_id)

    def _ensure_started(self, id: str) -> None:
        self.start(id, None)

    def success(self, id: str) -> None:
        self._ensure_started(id)
        task_id = self._map[id]
        self._job_ctl.update(task_id, completed=sys.float_info.max)
        self._todo.discard(id)

    def skip(self, id: str, reason: str) -> None:
        self._ensure_started(id)
        task_id = self._map[id]
        details = "skipped"
        if reason:
            details += f": {reason}"
        self._job_ctl.update(task_id, completed=sys.float_info.max, details=details)
        self._todo.discard(id)

    def error(self, id: str, details: str):
        self._ensure_started(id)
        task_id = self._map[id]
        if not details:
            details = ""
        self._job_ctl.update(task_id, details=details)
        self._job_ctl.stop_task(task_id)
        # self._job_ctl._tasks[task_id].finished = True
        self._todo.discard(id)
        self._job_ctl.stop

    def close(self) -> None:
        self._todo.clear()
        self._closed = True
        self._view_exited.wait()

    def show(self) -> None:
        try:
            with Live(self._job_ctl, refresh_per_second=10):
                self._view_started.set()
                while True:
                    self.refresh()
                    if self._closed:
                        break
                    if self._auto_close and not self._todo:
                        break
                    sleep(0.2)
        finally:
            self._view_exited.set()

    @staticmethod
    def create_async(auto_close: bool = False) -> "JobView":
        inst = JobView(auto_close)
        t = threading.Thread(target=inst.show, daemon=False)
        t.start()
        inst._view_started.wait()
        return inst



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import logging
from http.client import HTTPResponse
from typing import Callable

import httpx
from authlib.integrations.httpx_client import OAuth2Client
from hcs_core.ctxp import jsondot

log = logging.getLogger(__name__)


def parse(resp: httpx.Response):
    if not resp.content:
        return None
    content_type = resp.headers["Content-Type"]
    if content_type.startswith("text"):
        return resp.text
    if len(resp.content) > 3:
        try:
            data = resp.json()
            if isinstance(data, list):
                return [jsondot.dotify(obj) for obj in data]
            elif isinstance(data, dict):
                return jsondot.dotify(data)
        except:
            log.info("--- Fail parsing json. Dump content ---")
            log.info(resp.content)
            raise
    else:
        return None


def _raise_on_4xx_5xx(response: httpx.Response):
    if not response.is_success:
        response.read()
        if len(response.text) > 0:
            text = _try_formatting_json(response.text)
            log.debug(text)

    response.raise_for_status()


def _try_formatting_json(text: str):
    try:
        return json.dumps(json.loads(text), indent=4)
    except:
        return text


def _log_request(request):
    if not log.isEnabledFor(logging.DEBUG):
        return
    log.debug("\n")
    log.debug(f"--> {request.method} {request.url}")
    log.debug(f"--> {request.headers}")

    if len(request.content) > 0:
        text = _try_formatting_json(request.content)
        log.debug(text)


def _log_response(response: httpx.Response):
    if not log.isEnabledFor(logging.DEBUG):
        return
    request = response.request
    log.debug(f"<-- {request.method} {request.url} - {response.status_code}")
    log.debug(f"<-- {request.headers}")
    response.read()
    if len(response.text) > 0:
        text = _try_formatting_json(response.text)
        log.debug(text)
    log.debug("\n")


def _parse_resp(resp: httpx.Response):
    if not resp.content:
        return
    content_type = resp.headers["Content-Type"]
    if content_type.startswith("text"):
        return resp.text
    if content_type == "application/json" and resp.content:
        try:
            data = resp.json()
            return jsondot.dotify(data)
        except:
            log.info("--- Fail parsing json. Dump content ---")
            log.info(resp.content)
            raise
    return resp.text


def _is_404(e: httpx.HTTPStatusError) -> bool:
    return e.response.status_code == 404


def on404ReturnNone(func):
    try:
        resp = func()
        return _parse_resp(resp)
    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            return None
        raise


class EzClient:
    def __init__(self, base_url: str, oauth_client: OAuth2Client = None, lazy_oauth_client: Callable = None) -> None:
        # self._client = httpx.Client(base_url=base_url, timeout=30, event_hooks=event_hooks)
        self._base_url = base_url
        self._client_impl = oauth_client
        self._lazy_oauth_client = lazy_oauth_client

    def _client(self):
        if not self._client_impl:
            client = self._lazy_oauth_client()
            client.base_url = self._base_url
            client.timeout = 30
            request_hooks = client.event_hooks["request"]
            response_hooks = client.event_hooks["response"]
            if _log_request not in request_hooks:
                request_hooks.append(_log_request)
            if _log_response not in response_hooks:
                response_hooks.append(_log_response)
            if _raise_on_4xx_5xx not in response_hooks:
                response_hooks.append(_raise_on_4xx_5xx)
            self._client_impl = client
        return self._client_impl

    def post(self, url: str, json: dict = None, text: str = None, files=None, headers: dict = None):
        resp = self._client().post(url, json=json, content=text, files=files, headers=headers)
        return _parse_resp(resp)

    def get(self, url: str, raise_on_404: bool = False):
        try:
            # print("->", url)
            resp = self._client().get(url)
            return _parse_resp(resp)
        except httpx.HTTPStatusError as e:
            if _is_404(e):
                if raise_on_404:
                    raise e
                else:
                    pass
            else:
                raise

    def patch(self, url: str, json: dict):
        resp = self._client().patch(url, json=json)
        return _parse_resp(resp)

    def delete(self, url: str, raise_on_404: bool = False):
        try:
            resp = self._client().delete(url)
            return _parse_resp(resp)
        except httpx.HTTPStatusError as e:
            if _is_404(e):
                if raise_on_404:
                    raise
                else:
                    pass
            else:
                raise

    def put(self, url: str, json: dict):
        resp = self._client().put(url, json=json)
        return _parse_resp(resp)

    def close(self):
        self._client().close()

    def dump_response(self, response: HTTPResponse):
        log.info("response text: " + response.text())



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import yaml
from typing import Tuple
import os
from hcs_core.ctxp.data_util import load_data_file, process_variables


class PlanException(Exception):
    pass


class PluginException(Exception):
    pass


class Blueprint:
    var: dict
    default: dict
    provider: dict
    resource: dict

    def __init__(self):
        self.var = {}
        self.default = {}
        self.provider = {}
        self.resource = {}

    def __repr__(self):
        return "Blueprint"


def load_files(files: list[str]) -> dict:
    ret = {}
    for file in files:
        data = load_data_file(file)
        if data is None or isinstance(data, str):
            raise FileNotFoundError("Fail loading file: " + file)
        ret = _merge_dict_fail_on_dup(ret, _smart_load_file(file))
    return ret


def process_template(template, use_env: bool = True) -> Tuple[dict, dict]:
    # Ensure default sections are not None
    if "var" not in template:
        template["var"] = {}
    if "default" not in template:
        template["default"] = {}
    if "resource" not in template:
        template["resource"] = {}
    if "runtime" not in template:
        template["runtime"] = {}
    if "provider" not in template:
        template["provider"] = {}

    _validate_blueprint(template)

    deployment_id = template["deploymentId"]
    if isinstance(deployment_id, int):
        template["deploymentId"] = str(deployment_id)
    elif isinstance(deployment_id, str):
        pass
    else:
        raise PlanException("Invalid deploymentId type. Expect string, actual: " + type(deployment_id).__name__)
    bp, pending = _materialize_blueprint(template, use_env)
    return bp, pending


def _validate_blueprint(blueprint: dict):
    _validate_required_keys(blueprint)
    _validate_resource_schema(blueprint)
    _validate_resource_id_not_conflict_to_reserved_names(blueprint)
    _validate_no_conflict_resource_id_provider_types_runtime_id(blueprint)
    _validate_statement_after(blueprint)


def _validate_required_keys(blueprint: dict):
    required_keys = ["deploymentId", "resource"]
    for k in required_keys:
        if k not in blueprint:
            raise PlanException("Missing required property in blueprint: " + k)


def _validate_resource_schema(blueprint: dict):
    resource = blueprint["resource"]
    if not resource:
        return
    required_keys = set(["kind"])
    optional_keys = set(["eta", "data", "conditions", "for", "after"])
    for k, v in resource.items():

        def _raise(reason):
            raise PlanException(f"Invalid blueprint: {reason}. Resource: {k}")

        actual_keys = set(v.keys())
        missed_keys = required_keys - actual_keys
        if missed_keys:
            _raise(f"Missing required keys: {missed_keys}")
        extra_keys = actual_keys - required_keys - optional_keys
        if extra_keys:
            _raise(f"Unknown extra keys: {extra_keys}")


def _get_duplicates(lst):
    return [item for item in set(lst) if lst.count(item) > 1]


def _validate_statement_after(blueprint: dict):
    def _raise(owner, reason):
        raise PlanException(f"Invalid statement: after. Owner={owner}, reason={reason}.")

    items = {}
    items |= blueprint["resource"]
    items |= blueprint["runtime"]

    def _validate_after(target_name, owner_resource_name):
        if not isinstance(target_name, str):
            _raise(owner_resource_name, f"Invalid value type: {type(target_name).__name__}")

        if target_name not in items:
            _raise(owner_resource_name, "Target not found: " + target_name)

    for k, v in items.items():
        after = v.get("after")
        if after:
            if isinstance(after, list):
                for a in after:
                    _validate_after(a, k)

                dup = _get_duplicates(after)
                if dup:
                    _raise(k, "Duplicated keys: " + dup)
            elif isinstance(after, str):
                _validate_after(after, k)
            else:
                _raise(k, "Invalid type, expect str or list, got: " + type(after).__name__)


def _validate_no_conflict_resource_id_provider_types_runtime_id(blueprint: dict):
    provider_names = set()
    runtime_names = set(blueprint["runtime"].keys())
    for v in blueprint["resource"].values():
        name, _ = v["kind"].split("/")
        provider_names.add(name)

    declared_provider_names = set(blueprint["provider"].keys())
    excessive = declared_provider_names - provider_names
    if excessive:
        raise PlanException(f"Invalid blueprint. Unused provider definition: {excessive}.")
    resource_names = set(blueprint["resource"].keys())
    conflict = provider_names & resource_names
    if conflict:
        raise PlanException(f"Invalid blueprint. Provider ID and resource ID conflict: {conflict}.")
    conflict = runtime_names & resource_names
    if conflict:
        raise PlanException(f"Invalid blueprint. Runtime ID and resource ID conflict: {conflict}.")
    conflict = provider_names & runtime_names
    if conflict:
        raise PlanException(f"Invalid blueprint. Provider ID and runtime ID conflict: {conflict}.")


def _validate_resource_id_not_conflict_to_reserved_names(blueprint: dict):
    reserved_names_for_state = ["result", "pending", "log", "destroy_output"]
    reserved_names_for_blueprint = ["default", "var", "env", "provider", "resource", "runtime"]
    reserved_names_for_function = ["profile", "context"]
    existing_names_top_level = blueprint.keys()
    reserved_names = set(
        [
            *existing_names_top_level,
            *reserved_names_for_blueprint,
            *reserved_names_for_state,
            *reserved_names_for_function,
        ]
    )
    for name in blueprint["resource"]:
        if name in reserved_names:
            raise PlanException("Invalid blueprint. Resource name conflicts to a reserved name: " + name)
    for name in blueprint["runtime"]:
        if name in reserved_names:
            raise PlanException("Invalid blueprint. Runtime name conflicts to a reserved name: " + name)
    for name in blueprint["provider"]:
        if name in reserved_names:
            raise PlanException("Invalid blueprint. Provider name conflicts to a reserved name: " + name)


def _smart_load_file(file: str):
    if not os.path.exists(file):
        raise Exception("File not found: " + file)
    if not os.path.isfile(file):
        raise Exception("Not a file: " + file)
    if file.endswith(".json"):
        with open(file, "r") as f:
            return json.load(f)
    elif file.endswith(".yaml") or file.endswith(".yml"):
        with open(file, "r") as f:
            return yaml.safe_load(f)
    else:
        raise Exception("Unknown file extention: " + file)


def _merge_dict_fail_on_dup(o1: dict, o2: dict) -> dict:
    ret = dict(o1)
    for k, v in o2.items():
        if k in o1:
            raise Exception("Fail processing file. Duplicated key found: " + k)
        ret[k] = v
    return ret


def _materialize_blueprint(template: dict, use_env: bool) -> Tuple[dict, dict]:
    ret = process_variables(template, None, use_env)
    return template, ret["pending"]



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Callable
import time
from hcs_core.ctxp import profile, panic, CtxpException
from hcs_core.sglib import hcs_client
from hcs_core.util.query_util import with_query, PageRequest
from hcs_core.util import duration
from hcs_core.util import exit


def hdc_service_client(service_name: str):
    url = profile.current().hcs.url
    if not url.endswith("/"):
        url += "/"
    url += service_name
    return hcs_client(url)


def _get_region_url(region_name: str):
    regions = profile.current().hcs.regions
    if not region_name:
        return regions[0].url
    for r in regions:
        if r.name.lower() == region_name.lower():
            return r.url
    names = []
    for r in regions:
        names.append(r.name)
    panic(f"Region not found: {region_name}. Available regions: {names}")


def regional_service_client(region_name: str, service_name: str):
    # 'https://dev1b-westus2-cp103a.azcp.horizon.vmware.com/vmhub'
    url = _get_region_url(region_name)
    if not url:
        panic("Missing profile property: hcs.regions")
    if not url.endswith("/"):
        url += "/"
    url += service_name
    return hcs_client(url)


def _with_org_id(url: str, org_id: str):
    if org_id:
        if url.find("?") < 0:
            url += "?"
        url += "org_id=" + org_id
    return url


class default_crud:
    def __init__(self, client, base_context: str, resource_type_name: str):
        self._client = client
        self._base_context = base_context
        self._resource_type_name = resource_type_name

    def get(self, id: str, org_id: str, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
        url = with_query(f"{self._base_context}/{id}", **kwargs)
        # print(url)
        return self._client.get(url)

    def list(self, org_id: str, fn_filter: Callable = None, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id

        def _get_page(query_string):
            url = self._base_context + "?" + query_string
            # print(url)
            return self._client.get(url)

        return PageRequest(_get_page, fn_filter, **kwargs).get()

    def create(self, payload: dict, headers: dict = None, **kwargs):
        url = with_query(f"{self._base_context}", **kwargs)
        # print(url)
        # import json
        # print(json.dumps(payload, indent=4))
        if isinstance(payload, str):
            return self._client.post(url, text=payload, headers=headers)
        if isinstance(payload, dict):
            return self._client.post(url, json=payload, headers=headers)
        return self._client.post(url, json=payload, headers=headers)

    def upload(self, files, **kwargs):
        url = with_query(f"{self._base_context}", **kwargs)
        return self._client.post(url, files=files)

    def delete(self, id: str, org_id: str, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
        url = with_query(f"{self._base_context}/{id}", **kwargs)
        # print(url)
        return self._client.delete(url)

    def wait_for_deleted(self, id: str, org_id: str, timeout: str, fn_is_error: Callable = None):
        name = self._resource_type_name + "/" + id
        fn_get = lambda: self.get(id, org_id)
        return wait_for_res_deleted(name, fn_get, timeout=timeout, fn_is_error=fn_is_error)

    def update(self, id: str, org_id: str, data: dict, **kwargs):
        if org_id:
            kwargs["org_id"] = org_id
        url = with_query(f"{self._base_context}/{id}")
        return self._client.patch(url, data)


def _parse_timeout(timeout: str):
    if isinstance(timeout, int):
        return timeout
    if isinstance(timeout, str):
        return duration.to_seconds(timeout)

    raise CtxpException(f"Invalid timout. Type={type(timeout).__name__}, value={timeout}")


def wait_for_res_deleted(
    resource_name: str,
    fn_get: Callable,
    timeout: str,
    polling_interval_seconds: int = 10,
    fn_is_error: Callable = None,
):
    timeout_seconds = _parse_timeout(timeout)
    start = time.time()
    while True:
        t = fn_get()
        if t is None:
            return
        if fn_is_error:
            if fn_is_error(t):
                msg = f"Failed deleting resource '{resource_name}', resource in Error state."
                raise CtxpException(msg)

        now = time.time()
        remaining_seconds = timeout_seconds - (now - start)
        if remaining_seconds < 1:
            msg = f"Timeout waiting for resource '{resource_name}' to be deleted."
            raise TimeoutError(msg)
        sleep_seconds = remaining_seconds
        if sleep_seconds > polling_interval_seconds:
            sleep_seconds = polling_interval_seconds
        exit.sleep(sleep_seconds)


def wait_for_res_status(
    resource_name: str,
    fn_get: Callable,
    get_status: Callable,
    status_map: dict = None,
    is_ready: Callable = None,
    is_error: Callable = None,
    is_transition: Callable = None,
    timeout: str = "10m",
    polling_interval: str = "20s",
    not_found_as_success: bool = False,
):
    timeout_seconds = _parse_timeout(timeout)
    polling_interval_seconds = _parse_timeout(polling_interval)
    if polling_interval_seconds < 3:
        polling_interval_seconds = 3
    start = time.time()
    prefix = f"Error waiting for resource {resource_name}: "

    if isinstance(get_status, str):
        field_name = get_status
        get_status = lambda t: t[field_name]
    if status_map:
        if isinstance(status_map["ready"], str):
            status_map["ready"] = [status_map["ready"]]
        if isinstance(status_map["transition"], str):
            status_map["transition"] = [status_map["transition"]]
        if isinstance(status_map["error"], str):
            status_map["error"] = [status_map["error"]]
        if is_ready:
            raise CtxpException("Can not specify is_ready when status_map is provided.")
        if is_error:
            raise CtxpException("Can not specify is_error when status_map is provided.")
        if is_transition:
            raise CtxpException("Can not specify is_transition when status_map is provided.")
        is_ready = lambda s: s in status_map["ready"]
        is_error = lambda s: s in status_map["error"]
        is_transition = lambda s: s in status_map["transition"]
    else:
        if not is_ready:
            raise CtxpException("Either status_map or is_ready must be specified.")
        if not is_error:
            raise CtxpException("Either status_map or is_error must be specified.")
        if not is_transition:
            raise CtxpException("Either status_map or is_transition must be specified.")

    while True:
        t = fn_get()
        if t is None:
            if not_found_as_success:
                return
            raise CtxpException(prefix + "Not found.")
        status = get_status(t)
        # print(f"DDD - status {status}")
        if is_error(status):
            msg = prefix + f"Status error. Actual={status}"
            if status_map:
                msg += f", expected={status_map['ready']}"
            raise CtxpException(msg)
        if is_ready(status):
            return t
        if not is_transition(status):
            raise CtxpException(
                prefix + f"Unexpected status: {status}. If this is a transition, add it to status_map['transition']."
            )

        now = time.time()
        remaining_seconds = timeout_seconds - (now - start)
        if remaining_seconds < 1:
            raise TimeoutError(prefix + "Timeout.")
        sleep_seconds = remaining_seconds
        if sleep_seconds > polling_interval_seconds:
            sleep_seconds = polling_interval_seconds

        # print(f"DDD sleeping {sleep_seconds}")
        exit.sleep(sleep_seconds)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import click
import sys
from click.core import Group
import os.path as path
import os
import re
import importlib
import importlib.util
from pathlib import Path
from .util import print_output, print_error, validate_error_return, avoid_trace_for_ctrl_c
from .extension import ensure_extension

_eager_loading = os.environ.get("_CTXP_EAGER_LOAD")
if _eager_loading:
    print("_CTXP_EAGER_LOAD:", _eager_loading, file=sys.stderr)


class LazyGroup(click.Group):
    def __init__(self, mod_path: Path = None, extension=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mod_path = mod_path
        self._lazy_commands_loaded = False
        self._extension = extension
        if _eager_loading:
            self._ensure_subcommands()

    def list_commands(self, ctx):
        self._ensure_subcommands()
        return super().list_commands(ctx)

    def get_command(self, ctx, cmd_name):
        self._ensure_subcommands()
        return super().get_command(ctx, cmd_name)

    def _ensure_subcommands(self):
        self._ensure_extension()
        if not self._lazy_commands_loaded and self.mod_path:
            _add_subcommands(self.mod_path.absolute(), self)
            self._lazy_commands_loaded = True

    def invoke(self, ctx):
        self._ensure_extension()
        return super().invoke(ctx)

    def _ensure_extension(self):
        if self._extension:
            ensure_extension(self._extension)


def _add_built_in_commands(group: Group):
    script_dir = path.abspath(path.join(path.dirname(path.realpath(__file__)), "."))
    build_in_cmds_dir = path.join(script_dir, "built_in_cmds")
    _add_subcommands(build_in_cmds_dir, group)


def _ensure_sub_group(current: Group, mod_path: Path):
    name = mod_path.name.replace("_", "-")
    meta = _read_group_meta(mod_path)
    help = meta.get("help")
    extension = meta.get("extension")

    subgroup = current.commands.get(name)
    if subgroup and isinstance(subgroup, Group):
        if isinstance(subgroup, LazyGroup):
            subgroup.mod_path = mod_path
        return subgroup

    subgroup = LazyGroup(extension=extension, name=name, help=help, mod_path=mod_path)
    current.add_command(subgroup)
    return subgroup


_property_pattern = re.compile("(.*)\\s*=\\s*(.*)$")


def _read_group_meta(mod_path: Path) -> dict:
    meta_file = mod_path.absolute().joinpath("__init__.py")

    ret = {"help": None, "hidden": False, "extension": None}
    if meta_file.exists():
        with open(meta_file, "r") as f:
            lines = f.readlines()
            for line in lines:
                m = _property_pattern.match(line)
                if m:
                    name = m.group(1).strip()
                    value = eval(m.group(2))
                    ret[name] = value
    return ret


_excluded_names = ["__pycache__", "__init__.py", ".DS_Store"]


def _add_subcommands(commands_dir: str, group: Group):
    for mod_path in Path(commands_dir).glob("*"):
        if mod_path.name in _excluded_names:
            continue
        if mod_path.is_dir():
            _ensure_sub_group(group, mod_path)
        elif mod_path.name.endswith(".py"):
            _import_cmd_file(mod_path, group)
        else:
            pass
            # print("Unrecognized sub cmd: " + mod_path.name)
    return


def _import_cmd_file(mod_path: Path, parent: click.core.Group):
    mod_name = f"{mod_path}"
    # print("Loading ---> ", mod_name)

    spec = importlib.util.spec_from_file_location(mod_name, mod_path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    # mod = importlib.import_module(mod_name)
    # filter out any things that aren't a click Command
    cli_methods = []
    for attr in dir(mod):
        foo = getattr(mod, attr)
        if callable(foo):
            cli_methods.append(foo)

    known_groups = []

    def is_in_known_group(c: click.core.Command):
        for g in known_groups:
            if g.get_command(None, c.name):
                return True

    # Find groups first
    for foo in cli_methods:
        if isinstance(foo, click.core.Group):
            parent.add_command(foo)
            known_groups.append(foo)

    for foo in cli_methods:
        if isinstance(foo, click.core.Group):
            continue

        if isinstance(foo, click.core.Command):
            foo = _default_io(foo)  # decorate with default parameters
            if is_in_known_group(foo):
                continue
            parent.add_command(foo)


# Create a new decorator that combines the individual decorators
def _default_io(cmd: click.Command):

    from .cli_options import output, field, id_only, first

    cmd = output(cmd)
    # cmd = cli_options.verbose(cmd)
    cmd = field(cmd)
    cmd = id_only(cmd)
    cmd = first(cmd)
    callback = cmd.callback

    def inner(*args, **kwargs):
        io_args = {
            "output": kwargs.pop("output"),
            #'verbose': kwargs.pop('verbose'),
            "field": kwargs.pop("field"),
            "id_only": kwargs.pop("id_only"),
            "first": kwargs.pop("first"),
        }
        if io_args["output"] == "table":

            def _format(data):
                if not hasattr(callback, "formatter"):
                    from .util import CtxpException

                    raise CtxpException("Table output is specified, but no custom formatter specified on the command.")
                formatter = callback.__getattribute__("formatter")
                return formatter(data)

            io_args["format"] = _format
        ret = callback(*args, **kwargs)
        return _process_result(ret, **io_args)

    cmd.callback = inner
    return cmd


def _process_result(result, **kwargs):
    if result is None:
        return
    if isinstance(result, tuple):
        data, return_code = result

        is_error = isinstance(data, Exception) or return_code is not None
        if is_error:
            if return_code is None:
                return_code = 1
            validate_error_return(data, return_code)
            if isinstance(data, Exception):
                print_error(data)
            else:
                print_output(data, kwargs, file=sys.stderr)
            ctx = click.get_current_context()
            ctx.exit(return_code)
        # else fall-through
    elif isinstance(result, Exception):
        print_error(result)
        ctx = click.get_current_context()
        ctx.exit(1)
    else:
        print_output(result, kwargs)


def init(main_cli: click.Group, commands_dir: str):
    avoid_trace_for_ctrl_c()
    _add_built_in_commands(main_cli)
    _add_subcommands(commands_dir, main_cli)
    # main_cli.result_callback()(_process_result)
    ret = main_cli(obj={})
    return ret



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import logging
from typing import Any
from . import jsondot
from .jsondot import dotdict
from . import profile
from . import context

log = logging.getLogger(__name__)

_variables = None


def _vars() -> dotdict:
    global _variables
    _variables = context.get(".variables")
    return _variables


def _save():
    context.set(".variables", _variables)


def set(name: str, value: Any):
    variables = _vars()
    if name in variables:
        existing = variables.get(name)
        if existing == value:
            return
        log.debug(f"Updating [{name}] from [{existing}] to [{value}]")
    else:
        log.debug(f"Setting [{name}] to [{value}]")
    variables[name] = value
    _save()


def delete(name: str):
    variables = _vars()
    if name in variables:
        del variables[name]
        _save()


def get(name: str):
    variables = _vars()
    v = None
    if name in variables:
        v = variables[name]
    else:
        plain_profile = _get_plain_profile()
        if name in plain_profile:
            v = plain_profile[name]
        else:
            raise Exception("Variable not found in context and profile: " + name)
    return v


def apply(template, additional_vars: dict = None):
    mapping = _vars().copy()
    if additional_vars:
        mapping |= additional_vars

    # merge profile
    mapping |= _get_plain_profile()

    # The Python default Template.substitute has too many limitations...
    def substitute(text: str):
        pattern = re.compile(".*\\${(.+?)}.*")
        while True:
            m = pattern.match(text)
            if not m:
                break
            name = m.group(1)
            val = str(get(name))
            text = text.replace("${" + name + "}", val)
        return text

    def apply_impl(target):
        t = type(target)
        try:
            if t is str:
                target = substitute(target)
            elif t is dict:
                for k, v in target.items():
                    target[k] = apply_impl(v)
            elif t is list:
                for i in range(len(target)):
                    target[i] = apply_impl(target[i])
            else:
                pass
            return target
        except KeyError as e:
            raise Exception("Variable not defined in context") from e


def _get_plain_profile():
    global _plain_profile
    if _plain_profile is None:
        _plain_profile = {}
        data = profile.plain()
        for k in data:
            _plain_profile["profile." + k] = data[k]
    return _plain_profile


def load_template(name: str):
    tmpl = jsondot.load(name)
    return apply(tmpl)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import click


verbose = click.option(
    "--verbose",
    "-v",
    count=True,
    default=0,
    hidden=True,
    help="Print debug logs",
)

output = click.option(
    "--output",
    "-o",
    type=click.Choice(["json", "json-compact", "yaml", "text", "table"]),
    default=None,
    hidden=True,
    help="Specify output format",
)

field = click.option(
    "--field",
    type=str,
    required=False,
    hidden=True,
    help="Specify fields to output, in comma separated field names.",
)

wait = click.option(
    "--wait",
    "-w",
    type=str,
    required=False,
    default="0",
    help="Wait time. E.g. '30s', or '5m'. Default: 10m. Specify '0' to disable waiting and return immediately.",
)

search = click.option(
    "--search",
    "-s",
    type=str,
    required=False,
    help="Specify REST search. E.g. 'name $eq something'. Note: use single quote in bash/sh.",
)

sort = click.option(
    "--sort",
    type=str,
    required=False,
    help="Ascending/Descending. Format is property,{asc|desc} and default is ascending",
)

limit = click.option(
    "--limit", "-l", type=int, required=False, default=100, help="Optionally, specify the number of records to fetch."
)

id_only = click.option(
    "--id-only",
    is_flag=True,
    type=bool,
    required=False,
    default=False,
    hidden=True,
    help="Return only id of the document",
)

first = click.option(
    "--first",
    is_flag=True,
    type=bool,
    required=False,
    default=False,
    hidden=True,
    help="Return only the first object, if the result is a list.",
)

force = click.option("--force/--grace", type=bool, default=True, help="Specify deletion mode: forceful, or graceful.")

confirm = click.option(
    "--confirm/--dry-run", "-y", type=bool, default=False, help="Confirm the operation without prompt."
)


def formatter(custom_fn):
    def decorator(f):
        f.formatter = custom_fn
        return f

    return decorator



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import base64
import os
import subprocess
from typing import List
from OpenSSL import crypto
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes


def base64_encode(text: str) -> str:
    return base64.b64encode(text.encode("ascii")).decode("ascii")


def base64_decode(encoded: str) -> str:
    return base64.b64decode(encoded.encode("ascii")).decode("ascii")


def decode_x509_pem(x509_cert_pem: str) -> x509.Certificate:
    x509_cert_pem = x509_cert_pem.strip()
    return x509.load_pem_x509_certificate(x509_cert_pem.encode("ascii"), default_backend())


def get_x509_subject(cert: x509.Certificate) -> dict:
    subject = {}
    for n in cert.subject:
        subject[n.oid._name] = n.value
    return subject


def split_cert_chain_pem(cert_chain_pem: str) -> list:
    cert_chain_pem = cert_chain_pem.strip()

    cert_chain_pem = cert_chain_pem.replace("TRUSTED CERTIFICATE", "CERTIFICATE")

    # Separators:
    # -----BEGIN CERTIFICATE-----
    # -----END CERTIFICATE-----
    # -----BEGIN TRUSTED CERTIFICATE-----
    # -----END TRUSTED CERTIFICATE-----

    dash = "-----"
    separator = dash + "\n" + dash
    parts = cert_chain_pem.split(separator)

    for i in range(len(parts)):
        if i > 0:
            parts[i] = dash + parts[i]
        if i < len(parts) - 1:
            parts[i] = parts[i] + dash
    return parts


def cert_chain_pem_to_certs(cert_chain_pem: str) -> List[x509.Certificate]:
    pems = split_cert_chain_pem(cert_chain_pem)
    return list(map(decode_x509_pem, pems))


def check_cert_equals(cert1: x509.Certificate, cert2: x509.Certificate) -> bool:
    subject1 = get_x509_subject(cert1)
    subject2 = get_x509_subject(cert2)
    if subject1 != subject2:
        print("Different subject:")
        print(subject1)
        print(subject2)
        return False
    if cert1.serial_number != cert2.serial_number:
        print("Different serial_number:")
        print(cert1.serial_number)
        print(cert2.serial_number)
        return False

    fingerprint1 = fingerprint_sha256(cert1)
    fingerprint2 = fingerprint_sha256(cert2)
    if fingerprint1 != fingerprint2:
        print("Different fingerprint:")
        print(fingerprint1)
        print(fingerprint2)
        return False
    if cert1.signature != cert2.signature:
        print("Different signature")
        return False
    return True


def fingerprint_sha256(cert) -> str:
    return bytearray(cert.fingerprint(hashes.SHA256())).hex()


def validate_cert_chain_pem_using_openssl(chain_pems: list):
    with open("temp_leaf.pem", "w") as f:
        f.write(chain_pems[0])
    with open("temp_chain.pem", "w") as f:
        for pem in chain_pems[1:]:
            f.write(pem)
            f.write("\n")
    cmd = "openssl verify -verbose -CAfile temp_chain.pem temp_leaf.pem"

    subprocess.run(cmd, shell=True, check=True, capture_output=False)

    os.unlink("temp_leaf.pem")
    os.unlink("temp_chain.pem")


# Generate pkey
def generate_key(type=crypto.TYPE_RSA, bits: int = 4096):
    key = crypto.PKey()
    key.generate_key(type, bits)
    return key


def _ascii(text: str) -> bytes:
    return text.encode("ascii")


def generate_CSR(nodename, sans=[], key_length: int = 2048):
    C = "US"
    ST = "California"
    L = "Palo Alto"
    O = "VMware, Inc."
    OU = "EUC"

    ss = []
    for i in sans:
        ss.append("DNS: %s" % i)
    ss = ", ".join(ss)

    req = crypto.X509Req()
    req.get_subject().CN = nodename
    req.get_subject().countryName = C
    req.get_subject().stateOrProvinceName = ST
    req.get_subject().localityName = L
    req.get_subject().organizationName = O
    req.get_subject().organizationalUnitName = OU
    # Add in extensions
    base_constraints = [
        crypto.X509Extension(_ascii("keyUsage"), False, _ascii("Digital Signature, Non Repudiation, Key Encipherment")),
        crypto.X509Extension(_ascii("basicConstraints"), False, _ascii("CA:FALSE")),
    ]
    x509_extensions = base_constraints
    # If there are SAN entries, append the base_constraints to include them.
    if ss:
        san_constraint = crypto.X509Extension("subjectAltName", False, ss)
        x509_extensions.append(san_constraint)
    req.add_extensions(x509_extensions)
    # Utilizes generateKey function to kick off key generation.
    key = generate_key(crypto.TYPE_RSA, key_length)
    req.set_pubkey(key)
    req.sign(key, "sha256")
    csr_pem = crypto.dump_certificate_request(crypto.FILETYPE_PEM, req).decode("ascii")
    private_key_pem = crypto.dump_privatekey(crypto.FILETYPE_PEM, key).decode("ascii")
    return csr_pem, private_key_pem



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .jsondot import dotdict
from .fstore import fstore

_store_impl: fstore = None


def init(config_path: str) -> None:
    global _store_impl
    _store_impl = fstore(config_path)


def get(name: str, reload: bool = False) -> dotdict:
    return _store_impl.get(name, reload)


def list() -> list:
    return _store_impl.keys()



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import click
import sys
import subprocess
from importlib import reload, import_module
import pkg_resources
from .util import CtxpException


def ensure_extension(required: str, parent_module_name: str = None):
    installed = {pkg.key for pkg in pkg_resources.working_set}  # pylint: disable=not-an-iterable
    if required in installed:
        return

    msg = f"This command requires an extension: {required}. Install now?"
    click.confirm(click.style(msg, fg="yellow"), default=True, abort=True)

    cmd = f"pip install {required}"
    try:
        msg = f"Executing: '{cmd}'"
        click.echo(click.style(msg, fg="yellow"))
        subprocess.run(cmd, stdout=sys.stdout, stderr=sys.stderr, text=True, shell=True, check=True)
    except Exception as e:
        error_message = f"Fail installing extensions. Error details: {e}"
        raise CtxpException(error_message)

    if parent_module_name:
        m = import_module(parent_module_name)
        reload(m)
    else:
        import_module(required.replace("-", "_"))
    reload(pkg_resources)



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .jsondot import dotdict
from .profile_store import profile_store, fstore


def _store() -> fstore:
    return profile_store("context")


def list() -> list:
    return _store().keys()


def get(name: str, reload: bool = False, default=None) -> dotdict:
    return _store().get(key=name, reload=reload, default=default)


def set(name: str, data: dict):
    return _store().save(name, data)


def delete(name: str):
    return _store().delete(name)


def file(name: str):
    return _store()._get_path(name)


def clear():
    return _store().clear()



"""
Copyright 2023-2023 VMware Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
    http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
import traceback
import logging
from hcs_core.ctxp.util import error_details

log = logging.getLogger(__name__)


class KopException(Exception):
    pass


class KopMode:
    create = "create"
    delete = "delete"
    update = "update"


class KopAction:
    start = "start"
    success = "success"
    error = "error"
    skip = "skip"


class _DummyJobView:
    @staticmethod
    def add(id, name):
        pass

    @staticmethod
    def update(id, text: str):
        pass

    @staticmethod
    def start(id, eta: str):
        pass

    @staticmethod
    def success(id):
        pass

    @staticmethod
    def skip(id, reason):
        pass

    @staticmethod
    def error(id, err):
        pass


_job_view = _DummyJobView


def attach_job_view(view):
    global _job_view
    _job_view = view


class KOP:
    MODE = KopMode

    def __init__(self, state: dict, kind: str, name: str, mode: str = KopMode.create) -> "KOP":
        self._state = state
        self._kind = kind
        self._name = name
        self._id = None
        self._mode = mode
        self._started = False
        self._closed = False
        job_id = kind + "/" + name
        _job_view.add(job_id, job_id)

    def _job_id(self):
        return self._kind + "/" + self._name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value:
            self.error(exc_value)
        else:
            # default closing upon leaving scope
            if not self._closed:
                self._success()

    def id(self, res_id: str):
        self._id = res_id
        _job_view.update(self._job_id(), res_id)

    def start(self, mode: str = None, eta: str = None):
        if self._started:
            raise KopException("Kop already started. This is a framework logging issue.")
        if self._closed:
            raise KopException("Kop already closed. This is a framework logging issue.")
        if mode:
            self._mode = mode
        self._started = True
        self._add_log(KopAction.start)
        _job_view.start(self._job_id(), eta)

    def _success(self):
        if not self._started:
            raise KopException(
                "Kop not started. This is a framework logging issue. Call kop.start() before starting the resource operation."
            )
        if self._closed:
            raise KopException("Kop already closed. This is a framework logging issue.")
        self._add_log(KopAction.success)
        self._closed = True
        _job_view.success(self._job_id())

    def error(self, err):
        if isinstance(err, Exception):
            if _need_stack_trace(err):
                traceback.print_exception(type(err), err, err.__traceback__)
            details = error_details(err)
        else:
            details = str(err)
        self._add_log(KopAction.error, details)
        self._closed = True
        _job_view.error(self._job_id(), details)

    def skip(self, reason):
        self._add_log(KopAction.skip, reason)
        self._closed = True
        _job_view.skip(self._job_id(), reason)

    def _add_log(self, action: str, details: str = None):
        labels = {
            KopMode.create: {
                KopAction.start: "[creating]",
                KopAction.success: "[created ]",
                KopAction.skip: "[skipped ]",
                KopAction.error: "[error   ]",
            },
            KopMode.delete: {
                KopAction.start: "[deleting]",
                KopAction.success: "[deleted ]",
                KopAction.skip: "[skipped ]",
                KopAction.error: "[error   ]",
            },
            KopMode.update: {
                KopAction.start: "[updating]",
                KopAction.success: "[updated ]",
                KopAction.skip: "[skipped ]",
                KopAction.error: "[error   ]",
            },
        }
        label_map = labels.get(self._mode)
        if not label_map:
            raise KopException(f"Invalid mode: {self._mode}")
        label = label_map[action]
        msg = f"{label} {self._kind}:{self._name}"
        if self._id:
            msg += " " + self._id

        if _job_view == _DummyJobView:
            log.info(msg)
        entry = {"name": self._name, "time": int(time.time()), "action": action, "id": self._id}
        if details:
            entry["details"] = details
        self._state["log"][self._mode].append(entry)
        if action == "error":
            if self._id:
                log.warning('Plugin "%s:%s" failed: %s. Id=%s', self._kind, self._name, details, self._id)
            else:
                log.warning('Plugin "%s:%s" failed: %s.', self._kind, self._name, details)


def _need_stack_trace(e):
    if isinstance(e, KeyError):
        return True
    if isinstance(e, TypeError):
        return True
